from qtpy.QtWidgets import QDialog, QVBoxLayout, QTabWidget
from qtpy.QtGui import QFont
from qtpy.QtCore import Signal
from qtdraw.core.group_view import GroupView


# ==================================================
class GroupTab(QDialog):
    textMessage = Signal(str)  # message
    selectedData = Signal(str, int)  # name, row
    deselectedData = Signal()
    rawDataChanged = Signal(str, int)  # name, row

    # ==================================================
    def __init__(self, ds, title="Group Tab", width=512, height=800, read_only=False, parent=None):
        super().__init__(parent)
        self.setWindowTitle(title)
        self.resize(width, height)

        self.tab = QTabWidget(self)
        bold = QFont()
        bold.setBold(True)
        self.tab.tabBar().setFont(bold)
        for name in ds.keys():
            tabItem = GroupView(ds, name, read_only, parent=self)
            tabItem.model.rawDataChanged.connect(self.rawDataChanged.emit)
            tabItem.textMessage.connect(self.textMessage.emit)
            tabItem.selectedData.connect(self.selectedData.emit)
            tabItem.deselectedData.connect(self.deselectedData.emit)
            self.tab.addTab(tabItem, name)
        self.tab.setUsesScrollButtons(True)

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.tab)

        self.setLayout(layout)

    # ==================================================
    def closeEvent(self, event):
        for i in range(self.tab.count()):
            self.tab.widget(i).itemDeselected()
        return super().closeEvent(event)



import os
import sys
from PyQt5.QtCore import QT_VERSION_STR
from qtpy.QtGui import QPixmap, QFont
from qtpy.QtWidgets import QDialog, QSizePolicy, QLabel, QGridLayout, QDialogButtonBox
import pyvista
import pyvistaqt
import pandas
import numpy
import sympy
import matplotlib
from gcoreutils.latex_util import latex_version


# ==================================================
class DialogAbout(QDialog):
    # ==================================================
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setModal(True)
        self.setWindowTitle("About")
        indent = 15

        self.resize(100, 100)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.sizePolicy().hasHeightForWidth())
        self.setSizePolicy(sizePolicy)

        self.icon = QLabel(self)
        pix = QPixmap(os.path.dirname(__file__) + "/qtdraw.png")
        pix = pix.scaledToHeight(int(2.5 * self.height()))
        self.icon.setPixmap(pix)

        self.lbl_qtdraw = QLabel("QtDraw", self)
        font = QFont()
        font.setBold(True)
        self.lbl_qtdraw.setFont(font)

        self.txt_copyright = QLabel(parent.copyright_str, self)

        self.txt_pyvista = QLabel(
            "\nPyVista: Ver. "
            + pyvista._version.__version__
            + "\n     with PyVistaQt: Ver. "
            + pyvistaqt.__version__
            + ", VTK: Ver. "
            + ".".join(map(str, pyvista.vtk_version_info)),
            self,
        )
        self.txt_pyvista.setIndent(indent)

        self.txt_pandas = QLabel("Pandas: Ver. " + pandas.__version__, self)
        self.txt_pandas.setIndent(indent)

        self.txt_numpy = QLabel("NumPy: Ver. " + numpy.__version__, self)
        self.txt_numpy.setIndent(indent)

        self.txt_sympy = QLabel("SymPy: Ver. " + sympy.__version__, self)
        self.txt_sympy.setIndent(indent)

        self.txt_qt = QLabel("PyQt: Ver. " + QT_VERSION_STR, self)
        self.txt_qt.setIndent(indent)

        self.txt_matplotlib = QLabel("Matplotlib: Ver. " + matplotlib.__version__ + "\n     with " + latex_version(), self)
        self.txt_matplotlib.setIndent(indent)

        ver = sys.version.replace("\n", "").replace("[", "\n    [")
        self.txt_python = QLabel("\nPython: Ver. " + ver, self)

        if parent._multipie_loaded:
            self.txt_multipie = QLabel("MultiPie: Ver. " + parent._multipie_loaded, self)
            self.txt_multipie.setIndent(indent)

        self.buttonBox = QDialogButtonBox(QDialogButtonBox.Ok)
        self.buttonBox.accepted.connect(self.accept)

        self.gridLayout = QGridLayout(self)
        self.gridLayout.setContentsMargins(20, 20, 20, 20)
        self.gridLayout.setHorizontalSpacing(30)
        self.gridLayout.setVerticalSpacing(0)

        self.gridLayout.addWidget(self.icon, 0, 0, 10, 1)
        self.gridLayout.addWidget(self.lbl_qtdraw, 0, 1, 1, 1)
        self.gridLayout.addWidget(self.txt_copyright, 1, 1, 1, 1)
        self.gridLayout.addWidget(self.txt_pyvista, 2, 1, 1, 1)
        self.gridLayout.addWidget(self.txt_qt, 3, 1, 1, 1)
        self.gridLayout.addWidget(self.txt_pandas, 4, 1, 1, 1)
        self.gridLayout.addWidget(self.txt_matplotlib, 5, 1, 1, 1)
        self.gridLayout.addWidget(self.txt_numpy, 6, 1, 1, 1)
        self.gridLayout.addWidget(self.txt_sympy, 7, 1, 1, 1)
        self.gridLayout.addWidget(self.txt_python, 9, 1, 1, 1)
        self.gridLayout.addWidget(self.buttonBox, 11, 0, 1, 2)

        if parent._multipie_loaded:
            self.gridLayout.addWidget(self.txt_multipie, 8, 1, 1, 1)



import sympy as sp
import numpy as np
from gcoreutils.nsarray import NSArray
from gcoreutils.list_util import list_to_table
from multipie.tag.tag_irrep import TagIrrep
from multipie.tag.tag_multipole import TagMultipole
from qtdraw.core.table_dialog import TableDialog
from qtdraw.multipie.setting import rcParams


# ==================================================
def show_group_info(group, name, header, role, align, data, vheader=None, width=1024, height=600, parent=None):
    no, _, IS, setting = group.tag.info()
    s = f"No.{no}: {str(group)}, {IS}"
    if setting:
        s += " (" + setting + " setting)"
    title = name + " - " + s

    table = TableDialog(
        data, title=title, header=header, vheader=vheader, role=role, align=align, width=width, height=height, parent=parent
    )
    table.show()


# ==================================================
def create_harmonics(group, rank, head, qtdraw, parent):
    pgh = group.harmonics
    head = head.replace("T", "Q").replace("M", "G")
    hs = pgh.select(rank=rank, head=head)
    n = len(hs)

    name = head + "_harmonics"
    header = ["symbol", "rank", "irrep.", "mul.", "comp.", "expression", "definition"]
    role = ["math", "text", "math", "text", "text", "math", "math"]
    align = ["center", "center", "center", "center", "center", "left", "left"]
    vheader = [str(i + 1) for i in range(n)]

    data = []
    for h in hs:
        mul = h.tag.mul
        comp = h.tag.comp
        irrep = h.tag.tag_irrep()
        dim = irrep.dim
        if mul < 1:
            mul = ""
        if dim == 1:
            comp = ""
        data.append(
            [h.latex(), rank, irrep.latex(), mul, comp, h.expression(v=NSArray.vector3d(head)).latex(), h.definition().latex()]
        )

    show_group_info(group, name, header, role, align, data, vheader=vheader, parent=parent)

    pos = NSArray([[np.cos(2 * np.pi * i / n), np.sin(2 * np.pi * i / n), 0.0] for i in range(n)], "vector", "value")
    color = rcParams["orbital_color_" + head]
    pname = qtdraw._get_name("orbital")
    for i in range(n):
        qtdraw.plot_orbital(
            pos[i],
            str(hs[i].expression(v=NSArray.vector3d())),
            size=0.8 / (rank + 1),
            color=color,
            name=pname,
            label=f"{i+1}",
            show_lbl=rcParams["show_label"],
        )
    qtdraw._plot_all_object()


# ==================================================
def create_harmonics_decomp(group, rank, head, to_pg, parent):
    pgh = group.harmonics
    to_pg = to_pg.split(".")[1][1:]  # remove space at top.
    head = head.replace("T", "Q").replace("M", "G")
    axial = head == "G"
    hs = pgh.select(rank=rank, head=head)
    decomp = group.irrep_decomposition(to_pg, rank, axial)
    n = len(hs)

    name = head + "_harmonics decomposition to " + to_pg
    header = ["symbol [" + str(group) + "]", "decomp. expression [" + to_pg + "]"]
    role = ["math", "math"]
    align = ["center", "center"]
    vheader = [str(i + 1) for i in range(n)]

    data = []
    for h, d in zip(hs, decomp):
        ex = sp.S(0)
        for c, b in d[1]:
            ex += c * TagMultipole(b).symbol()
        data.append([h.latex(), sp.latex(ex)])

    show_group_info(group, name, header, role, align, data, vheader=vheader, parent=parent)


# ==================================================
def create_character_table(group, parent):
    name = "character"
    ch = group.character
    n = len(ch.symmetry_operation())
    header = ["irrep."] + [f"SO{i+1}" for i in range(n)]
    role = ["math"] * (n + 1)
    align = ["center"] * (n + 1)

    data = []
    data.append([""] + [i.latex() for i in ch.symmetry_operation()])
    for r in ch.irrep_list:
        data.append([r.latex()] + [sp.latex(j) for j in ch.character(r)])

    show_group_info(group, name, header, role, align, data, parent=parent)


# ==================================================
def create_wyckoff(group, parent):
    name = "wyckoff_position"
    wp = group.wyckoff
    so = group.symmetry_operation
    n = list(wp.keys())[-1].n
    if not group.tag.is_point_group():
        ns = len(so.plus_set)
        n = n // ns
    n = min(10, n)
    header = ["position"] + [str(i + 1) for i in range(n)]
    role = ["text"] + ["math"] * n
    align = ["center"] * (n + 1)

    data = []
    if not group.tag.is_point_group():
        ps = ["+" + i for i in so.plus_set.latex()]
        data.append(["plus set"] + ps + [""] * (n - len(ps)))
        data.append([""] * (n - len(ps) + 2))

    for w in wp.keys():
        row = list_to_table(wp.position(w).latex(), n, "")
        for i, r in enumerate(row):
            if i == 0:
                data.append([str(w)] + r)
            else:
                data.append([""] + r)

    show_group_info(group, name, header, role, align, data, parent=parent)


# ==================================================
def create_product_table(group, parent):
    name = "product_table"
    so = group.symmetry_operation
    op = [i.latex() for i in so.keys()]
    header = ["SO"] + [f"SO{i+1}" for i in range(len(op))]
    role = ["math"] * (len(op) + 1)
    align = ["center"] * (len(op) + 1)
    vheader = [""] + [str(i + 1) for i in range(len(so))]

    data = []
    data.append([""] + op)
    for i in so.keys():
        data.append([i.latex()] + [so.product(i, j).latex() for j in so.keys()])

    show_group_info(group, name, header, role, align, data, vheader=vheader, parent=parent)


# ==================================================
def create_symmetry_operation(group, parent):
    so = group.symmetry_operation

    name = "symmetry_operation"
    header = ["symbol", "polar vector", "axial vector", "det"]
    role = ["math", "math", "math", "math"]
    align = ["center", "center", "center", "center"]
    vheader = [str(i + 1) for i in range(len(so))]

    data = []
    if not group.tag.is_point_group():
        ps = ["+" + i for i in so.plus_set.latex()]
        data.append(ps + [""] * (4 - len(ps)))
        data.append(["", "", "", ""])
        vheader = ["plus set", ""] + vheader

    pso = so.mat(axial=False)
    aso = so.mat(axial=True)
    for no in range(len(so)):
        p = pso[no][0:3, :]
        a = aso[no][0:3, :]
        data.append([so.full[no].latex(), p.latex(), a.latex(), p[0:3, 0:3].det().latex()])

    show_group_info(group, name, header, role, align, data, vheader=vheader, parent=parent)


# ==================================================
def create_v_cluster(group, wp, bond, qtdraw, parent):
    basis, site = group.virtual_cluster_basis(wyckoff=wp)
    n = min(10, len(site))

    name = "virtual cluster"
    header = ["symbol"] + [f"S{i+1}" for i in range(n)]
    role = ["math"] * (n + 1)
    align = ["center"] * (n + 1)

    data = []
    row = list_to_table(site.latex(), n, "")
    for i, r in enumerate(row):
        if i == 0:
            data.append([r"\text{site (r.c.)}"] + r)
        else:
            data.append([""] + r)
    for tag, v in basis.items():
        row = list_to_table(NSArray(v.tolist(), "scalar").latex(), n, "")
        for i, r in enumerate(row):
            if i == 0:
                t = tag.latex()
            else:
                t = ""
            data.append([t] + r)

    show_group_info(group, name, header, role, align, data, parent=parent)

    pname = qtdraw._get_name("site")
    color = rcParams["site_color"]
    for i in range(len(site)):
        qtdraw.plot_site(site[i], size=1.0, color=color, name=pname, label=f"{i+1}", show_lbl=rcParams["show_label"])

    if bond == "":
        bond = "1"
    bond = list(map(int, bond.split(" ")))
    d = NSArray.distance(site, site, qtdraw._G[0:3, 0:3])
    dkey = list(d.keys())
    for i in bond:
        name = f"b{i}"
        if i < len(d):
            for idxs in d[dkey[i]]:
                t, h = site[idxs[0]], site[idxs[1]]
                c = (t + h) / 2
                v = h - t
                qtdraw.plot_bond(c, v, name=name)

    qtdraw._plot_all_object()


# ==================================================
def create_response(group, rank, i_type, t_type, parent):
    dic = {("polar", "E"): "Q", ("polar", "M"): "T", ("axial", "E"): "G", ("axial", "M"): "M"}
    pgr = group.response
    head = dic[(i_type, t_type)]
    rt = pgr.select(rank=rank, head=head)

    name = i_type + "_response"
    header = ["symbol", "tensor"]
    role = ["math", "math"]
    align = ["center", "center"]

    data = []
    for tag in rt.keys():
        M = rt[tag]
        data.append([tag.latex(), M.latex()])

    data.append(["", ""])
    data.append([r"\text{component}", r"\text{multipole}"])
    for tag in rt.keys():
        def_mul = rt.definition[tag]
        for t, d in def_mul.items():
            data.append([sp.latex(t), sp.latex(d)])

    show_group_info(group, name, header, role, align, data, parent=parent)


# ==================================================
def create_atomic_mp(group, bra, ket, am, parent):
    name = "atomic multipole"
    header = ["symbol", "rank", "irrep.", "s", "k", "basis"]
    role = ["math", "text", "math", "text", "text", "math"]
    align = ["center"] * 6
    vheader = ["", ""] + [str(i + 1) for i in range(am.active_num())]

    data = []
    s = r"\langle " + str(bra) + r"|" + str(ket) + r"\rangle"
    data.append([""] * 5 + [s])
    data.append([""] * 6)
    for am1 in am.values():
        for tag, m in am1.items():
            data.append([tag.latex(), tag.rank, TagIrrep(tag.irrep).latex(), tag.s, tag.k, NSArray(m.tolist(), "matrix").latex()])

    show_group_info(group, name, header, role, align, data, vheader=vheader, parent=parent)



import numpy as np
from pymatgen.io.cif import CifParser
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.analysis.local_env import MinimumDistanceNN
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core.periodic_table import Element
from qtdraw.parser.element import element_color
from qtdraw.core.setting import rcParams
from qtdraw.parser.vesta import parse_vesta, create_vesta_graph

color_scheme = rcParams["plotter.color_scheme"]
digit = 6


# ==================================================
def get_graph(file):
    """
    get symmetrized structure graph.

    Args:
        file (str): file name.

    Returns:
        StructureGraph: pymatgen StructureGraph object.
    """
    parser = CifParser(file)
    structure = parser.get_structures(primitive=False, symmetrized=False)[0]
    sga = SpacegroupAnalyzer(structure)
    symmetrized = sga.get_symmetrized_structure()
    env = MinimumDistanceNN()
    graph = StructureGraph.with_local_env_strategy(symmetrized, env)

    return graph


# ==================================================
def get_model_cell(graph):
    """
    get model and cell.

    Args:
        graph (StructureGraph): pymatgen StructureGraph object.

    Returns: tuple.
        - str: name of model.
        - dict: unit-cell info.
        - float: unit-cell volume.
    """
    name = graph.structure.composition.reduced_formula

    lat = graph.structure.lattice
    cell = {
        "a": np.round(lat.a, digit),
        "b": np.round(lat.b, digit),
        "c": np.round(lat.c, digit),
        "alpha": np.round(lat.alpha, digit),
        "beta": np.round(lat.beta, digit),
        "gamma": np.round(lat.gamma, digit),
    }
    volume = lat.volume

    return name, cell, volume


# ==================================================
def get_site_info(graph):
    """
    get site information.

    Args:
        graph (StructureGraph): pymatgen StructureGraph object.

    Returns:
        list: site_info. (name, label, element, frac_coords, radius, color).
    """
    structure = graph.structure
    eq_sites = structure.equivalent_sites

    # grouping equivalent sites.
    dsites = {}
    for es in eq_sites:
        el = es[0].specie
        if type(el) != type(Element):
            el = str(el).replace("0+", "")
        else:
            el = el.symbol
        dsites[el] = dsites.get(el, []) + [es]

    site_info = []
    for name, v in dsites.items():
        for el_no, sl in enumerate(v):
            el_name = name + str(el_no + 1)
            for s_no, s in enumerate(sl):
                s_name = el_name + "_" + str(s_no + 1)
                frac_coords = s.frac_coords.round(digit)
                if name in element_color[color_scheme].keys():
                    radius = float(s.specie.atomic_radius)
                    color = element_color[color_scheme][name]
                else:
                    radius = 1.0
                    color = "silver"
                site_info.append((el_name, s_name, name, frac_coords, radius, color))

    return site_info


# ==================================================
def get_bond_info(graph, site_info):
    """
    get bond information.

    Args:
        graph (StructureGraph): pymatgen StructureGraph object.
        site_info (list): site info.

    Returns:
        list: bond_info. (name, label, center, vector, tail_color, head_color).
    """
    adjacency = graph.as_dict()["graphs"]["adjacency"]

    dbonds = {}
    for tail, tail_adjacency in zip(site_info, adjacency):
        t_name, _, _, t_pos, _, t_c = tail
        for head_info in tail_adjacency:
            h_name, _, _, h_pos, _, h_c = site_info[head_info["id"]]
            h_pos = np.array(head_info["to_jimage"]) + h_pos
            center = ((t_pos + h_pos) / 2).round(digit)
            vector = (h_pos - t_pos).round(digit)
            key = t_name + "-" + h_name
            dbonds[key] = dbonds.get(key, []) + [(center, vector, t_c, h_c)]

    bond_info = []
    for name, v in dbonds.items():
        for no, (center, vector, t_c, h_c) in enumerate(v):
            label = name + "_" + str(no + 1)
            bond_info.append((name, label, center, vector, t_c, h_c))

    return bond_info


# ==================================================
def plot_site_bond(qtdraw, name, cell, volume, site_info, bond_info):
    """
    plot site and bond.

    Args:
        name (str): name of model.
        cell (dict): unit-cell info.
        volume (float): unit-cell volume.
        site_info (list): site info.
        bond_info (list): bond info.
    """
    qtdraw._init_setting(model=name, cell=cell, axis_type="abc", clip=False)
    qtdraw._remove_all_actor()
    qtdraw._init_all()

    scale = volume ** (1 / 3)
    site_scale = rcParams["plotter.site_scale"]
    bond_scale = rcParams["plotter.bond_scale"]

    # plot sites.
    for name, label, _, frac_coords, radius, color in site_info:
        qtdraw.plot_site(str(frac_coords.tolist()), name=name, label=label, color=color, size=site_scale * radius * scale)

    # plot bonds.
    for name, label, center, vector, tail_color, head_color in bond_info:
        qtdraw.plot_bond(
            str(center.tolist()),
            str(vector.tolist()),
            color=tail_color,
            color2=head_color,
            name=name,
            label=label,
            width=bond_scale * scale,
        )

    qtdraw._plot_all_object()
    qtdraw.set_view()


# ==================================================
def plot_cif(qtdraw, filename):
    """
    read and plot CIF file.

    Args:
        filename (str): filename.
    """
    graph = get_graph(filename)
    name, cell, volume = get_model_cell(graph)
    site_info = get_site_info(graph)
    bond_info = get_bond_info(graph, site_info)
    plot_site_bond(qtdraw, name, cell, volume, site_info, bond_info)


# ==================================================
def plot_vesta(qtdraw, filename):
    """
    read and plot VESTA file.

    Args:
        filename (str): filename.
    """
    vesta_dict = parse_vesta(filename)
    graph = create_vesta_graph(vesta_dict)

    name, cell, volume = get_model_cell(graph)
    site_info = get_site_info(graph)
    bond_info = get_bond_info(graph, site_info)
    plot_site_bond(qtdraw, name, cell, volume, site_info, bond_info)



from qtpy.QtGui import QPixmap, QIcon
from qtpy.QtWidgets import (
    QWidget,
    QLabel,
    QCheckBox,
    QHBoxLayout,
    QBoxLayout,
    QButtonGroup,
    QRadioButton,
    QComboBox,
    QFrame,
    QSizePolicy,
)
from gcoreutils.convert_util import text_to_list, text_to_sympy, sympy_to_latex
from gcoreutils.nsarray import NSArray
from qtdraw.core.color_palette import color2pixmap, _color2pixmap
from qtdraw.core.pixmap_converter import latex2pixmap
from qtdraw.core.line_edit import LineEditor, Label


# ==================================================
class EditableWidget(QWidget):
    def __init__(self, label=None, callback=None, parent=None):
        super().__init__(parent)
        if callback is None:
            callback = self.default_callback

        self.editor = None
        self.inedit = False
        self.callback = callback

        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(15, 3, 15, 3)
        self.layout.setSpacing(3)

        if label is not None:
            self.label = Label(parent=self)
            self.setLabel(label)
            self.layout.addWidget(self.label)

        self.setLayout(self.layout)

    # ==================================================
    def set_editor(self, label="", init="", validator=None, decimal=4, check_var=None, callback=None):
        if label is None:
            self.editor = None
        else:
            self.editor = LineEditor(label, init, validator, decimal, check_var, callback, self)
            self.editor.setWindowTitle("Input Panel")
            self.editor.layout.setContentsMargins(3, 3, 3, 3)
            self.editor.resize(600, 2 * self.label.height())
            self.editor.hide()
            self.editor.closeEdit.connect(self.close_edit)
            self.editor.acceptEdit.connect(self.accept_edit)

    # ==================================================
    def setLabel(self, label):
        self.label.setText(label)

    # ==================================================
    def mouseDoubleClickEvent(self, event):
        if self.editor is not None and not self.inedit:
            self.inedit = True
            self.label.hide()
            self.editor.show()
            self.editor.editor.setFocus()

    # ==================================================
    def close_edit(self):
        self.editor.hide()
        self.label.show()
        self.inedit = False

    # ==================================================
    def accept_edit(self):
        self.setLabel(self.editor.text())
        self.close_edit()

    # ==================================================
    def default_callback(self, text):
        print(text)
        return True

    # ==================================================
    def __str__(self):
        return "Widget: " + self.label.text()


# ==================================================
class QtColorSelector(EditableWidget):
    """
    color selector widget.
    """

    # ==================================================
    def __init__(self, current="", color_type="color", callback=None, parent=None):
        """
        initialize the class.

        Args:
            current (str, optional): default color.
            color_type (str, optional): color/colormap/color_both
            callback (function, optional): callback function to get color name, f(name).
            parent (QWidget, optional): parent.

        Notes:
            - callback is f(colorname).
        """
        super().__init__(callback=callback, parent=parent)

        size = self.font().pointSize()

        color_pixmap, separator = color2pixmap(color_type, size)
        names = list(color_pixmap.keys())

        if current == "" or current is None:
            current = names[0]

        try:
            current = names.index(current)
        except ValueError:
            current = 0

        combo = QComboBox(self)
        combo.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        for s, c in color_pixmap.items():
            combo.addItem(QIcon(c), s, self)
        combo.setCurrentIndex(current)

        combo.setIconSize(next(iter(color_pixmap.values())).size())

        for i in separator:
            combo.insertSeparator(i)

        self.layout.addWidget(combo)

        def _callback(txt):
            self.callback(txt)
            return True

        combo.currentTextChanged.connect(_callback)


# ==================================================
class QtCheckBox(EditableWidget):
    """
    check box widget.
    """

    # ==================================================
    def __init__(self, label=None, current=True, callback_check=None, callback_label=None, read_only=False, parent=None):
        """
        initialize the class.

        Args:
            label (str, optional): label for check box.
            current (bool, optional): default status.
            callback_check (function, optional): callback function to get check status.
            callback_label (function, optional): callback function to get label.
            read_only (bool, optional): read only ?
            parent (QWidget, optional): parent.

        Notes:
            - callback is f(flag). If label makes editable, two callbacks f1(flag) and f2(text) are passed.
        """
        if callback_check is None:
            callback_check = self.default_check

        super().__init__(label, parent=parent)
        if callback_label is not None and label is not None and not read_only:
            self.set_editor(init=label, validator="latex", callback=callback_label)
            self.editor.setContentsMargins(3, 1, 0, 1)

        chkbox = QCheckBox("", self)
        sizePolicy = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        chkbox.setSizePolicy(sizePolicy)

        if current:
            chkbox.setChecked(True)
        else:
            chkbox.setChecked(False)

        if label is None:
            self.layout.addWidget(chkbox)
        else:
            self.layout.insertWidget(0, chkbox)
            self.label.setContentsMargins(13, 1, 0, 1)

        chkbox.toggled.connect(callback_check)

    # ==================================================
    def default_check(self, flag):
        print(flag)


# ==================================================
class QtRadioGroup(EditableWidget):
    """
    radio group widget.
    """

    # ==================================================
    def __init__(self, lst, current=0, vertical=False, callback=None, parent=None):
        """
        initialize the class.

        Args:
            lst (list): radio group text.
            current (int, optional): default index.
            vertical (bool, optional): align vertically ?
            callback (function, optional): callback function to get selected text, f(text).
            parent (QWidget, optional): parent.

        Notes:
            - callback is f(lst_index).
        """
        self.lst = lst

        super().__init__(callback=callback, parent=parent)

        if vertical:
            self.layout.setDirection(QBoxLayout.TopToBottom)

        self.select = QButtonGroup(self)
        for no, i in enumerate(lst):
            btn = QRadioButton(i, self)
            self.select.addButton(btn)
            if no == current:
                btn.setChecked(True)
            self.layout.addWidget(btn)

        def _callback(btn):
            no = self.lst.index(btn.text())
            self.callback(no)
            return True

        self.select.buttonClicked.connect(_callback)


# ==================================================
class QtComboBox(EditableWidget):
    """
    combo box widget.
    """

    # ==================================================
    def __init__(self, lst, current=0, callback=None, parent=None):
        """
        initialize the class.

        Args:
            lst (list): combo box text.
            current (int, optional): default index.
            callback (function, optional): callback function to get selected text, f(text).
            parent (QWidget, optional): parent.

        Notes:
            - callback is f(lst_index).
        """
        self.lst = lst

        super().__init__(callback=callback, parent=parent)

        combo = QComboBox(self)
        for s in lst:
            combo.addItem(s, self)
        combo.setCurrentIndex(current)

        self.layout.addWidget(combo)

        def _callback(txt):
            no = self.lst.index(txt)
            self.callback(no)
            return True

        combo.currentTextChanged.connect(_callback)


# ==================================================
class QtImage(EditableWidget):
    """
    image widget.
    """

    # ==================================================
    def __init__(self, label, width=100, callback=None, read_only=False, parent=None):
        """
        initialize the class.

        Args:
            label (str): file name of image.
            width (int, optional): width.
            callback (function, optional): callback when editable=True.
            read_only (bool, optional): read only ?
            parent (QWidget, optional): parent.

        Notes:
            - callback is f(filename).
        """
        self.w = width
        super().__init__(label, callback=callback, parent=parent)
        if not read_only:
            self.set_editor(init=label)

    # ==================================================
    def setLabel(self, label):
        # file existing check for label.
        pix = QPixmap(label)
        pix = pix.scaledToWidth(self.w)

        self.label.setPixmap(pix)


# ==================================================
class QtColorLabel(EditableWidget):
    """
    color label widget.
    """

    # ==================================================
    def __init__(self, label, parent=None):
        """
        initialize the class.

        Args:
            label (str): color or colormap name.
            parent (QWidget, optional): parent.
        """
        if label == "" or label is None:
            return

        super().__init__(label, parent=parent)

        size = self.font().pointSize()

        colorbox = _color2pixmap(label, "color_both", size)

        icon = QLabel(self)
        icon.setPixmap(colorbox)
        icon.setFrameStyle(QFrame.Box)
        icon.setFixedSize(colorbox.size())

        self.layout.insertWidget(0, icon)
        self.label.setContentsMargins(3, 1, 0, 1)


# ==================================================
class QtText(EditableWidget):
    """
    text widget.
    """

    # ==================================================
    def __init__(self, label, callback=None, validator=None, decimal=4, check_var=None, read_only=False, align=None, parent=None):
        """
        initialize the class.

        Args:
            label (str): text.
            callback (function, optional): callback function to get input, f(input).
            validator (str, optional): validator.
            decimal (int, optional): digit for floating number.
            check_var (list, optional): acceptable variables.
            read_only (bool, optional): read only ?
            align (str, optional): alignment, "left/center/right".
            parent (QWidget, optional): parent.
        """
        if align is None:
            align = "left"
        label = str(label).replace("'", "")
        self.validator = validator
        super().__init__(label, parent=parent)

        if callback is not None and not read_only:
            self.set_editor(init=label, validator=validator, decimal=decimal, check_var=check_var, callback=callback)
            self.setLabel(self.editor.text(), align)
        else:
            self.setLabel(label, align)

    # ==================================================
    def setLabel(self, label, align=None):
        if self.validator in [
            "s_row",
            "s_vector",
            "i_row",
            "i_pos_row",
            "i_vector",
            "i_pos_vector",
            "r_row",
            "r_pos_row",
            "r_vector",
            "r_pos_vector",
        ]:
            text = text_to_list(label)
            text = "[" + ", ".join(map(lambda i: str(i).strip(" "), text)) + "]"
        elif self.validator in [
            "s_column",
            "i_column",
            "i_pos_column",
            "r_column",
            "r_pos_column",
            "s_column_vector",
            "i_column_vector",
            "i_pos_column_vector",
            "r_column_vector",
            "r_pos_column_vector",
        ]:
            text = text_to_list(label)
            text = "\n".join(map(lambda i: str(i).strip(" "), text))
        elif self.validator in [
            "s_matrix",
            "s_vector_list",
            "i_matrix",
            "i_pos_matrix",
            "i_vector_list",
            "i_pos_vector_list",
            "r_matrix",
            "r_pos_matrix",
            "r_vector_list",
            "r_pos_vector_list",
        ]:
            text = text_to_list(label)
            text = "\n".join(["[" + ", ".join(map(lambda j: str(j).strip(" "), i)) + "]" for i in text])
        else:
            text = label

        self.label.setText(text)
        self.label.setAlignment(align)


# ==================================================
class QtMath(EditableWidget):
    """
    math widget.

    References:
        - Render TeX to SVG in Python using matplotlib and display with PyQt
        - https://gist.github.com/gmarull/dcc8218385014559c1ca46047457c364
    """

    # ==================================================
    def __init__(
        self,
        label,
        callback=None,
        validator=None,
        check_var=None,
        color="black",
        style="svg",
        dpi=300,
        read_only=False,
        align=None,
        parent=None,
    ):
        """
        initialize the class.

        Args:
            label (str or list): that can be conveted to sympy.
            callback (function, optional): callback function to get input, f(input).
            validator (str, optional): scalar/row/column/matrix.
            check_var (list, optional): acceptable variables.
            color (str, optional): color name.
            style (str, optional): pixmap style, "svg/png/jpg".
            dpi (int, optional): DPI of image.
            read_only (bool, optional): read only ?
            align (str, optional): alignment, "left/center/right".
            parent (QWidget, optional): parent.
        """
        if align is None:
            align = "left"
        if label is None:
            return
        if validator not in [None, "s_scalar", "s_row", "s_column", "s_matrix", "s_vector", "s_column_vector", "s_vector_list"]:
            return

        self.validator = validator
        self.color = color
        self.style = style
        self.dpi = dpi

        super().__init__(label, parent=parent)

        if callback is not None and not read_only:
            self.set_editor(init=label, validator=validator, check_var=check_var, callback=callback)
            self.setLabel(self.editor.text(), align)
        else:
            self.setLabel(label, align)

    # ==================================================
    def setLabel(self, label, align=None):
        if label == "":
            return
        if self.validator is not None:
            # latex = sympy_to_latex(text_to_sympy(label))
            latex = sympy_to_latex(NSArray(label).semi_evalf().tolist())
        else:
            latex = label

        size = self.font().pointSize()
        sp = r"\,\,\," if self.style == "svg" else ""

        if self.validator in ["s_row", "s_vector"]:
            latex = r"\begin{pmatrix}" + (r"&" + sp).join(latex) + r"\end{pmatrix}"
        elif self.validator in ["s_column", "s_column_vector"]:
            latex = (r"\begin{pmatrix}" + r"&".join(latex) + r"\end{pmatrix}").replace(r"&", r"\\")
        elif self.validator == "s_matrix":
            latex = r"\begin{pmatrix}" + r"\\".join([(r"&" + sp).join(i) for i in latex]) + r"\end{pmatrix}"
        elif self.validator == "s_vector_list":
            latex = (
                r"\begin{matrix}"
                + r"\\".join([r"\begin{pmatrix}" + (r"&" + sp).join(i) + r"\end{pmatrix}" for i in latex])
                + r"\end{matrix}"
            )

        pixmap = latex2pixmap(latex, size, self.color, self.dpi, self.style)
        if pixmap:
            self.label.setPixmap(pixmap)
            self.label.setAlignment(align)



import numpy as np
import vtk
from vtk import vtkParametricSpline
import pyvista as pv
from pyvista.utilities import surface_from_para
from qtpy.QtWidgets import QApplication, QStyleFactory
from gcoreutils.nsarray import NSArray
from gcoreutils.latex_util import latex_setting
from qtdraw.core.setting import rcParams
from qtdraw.core.color_palette import custom_colormap, all_colors

CHOP = 1e-4


# ==================================================
def create_application(style="fusion", font="Helvetica Neue", font_size=13, latex_mode="standard"):
    """
    create QApplication with style.

    Args:
        style (str, optional): GUI style, "fusion/windows/macintosh".
        font (str, optional): text font, "Helvetica Neue/Monaco/Osaka/Times New Roman".
        font_size (int, optional): text font size.
        latex_mode (str, optional): latex mode, "standard/times/pgf"

    Notes:
        - see, https://qiita.com/Hanjin_Liu/items/9df684920727f8a784c4
    """
    latex_setting(latex_mode)
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    global APPLICATION
    APPLICATION = app
    app.setStyle(QStyleFactory.create(style))
    style_sheet = "*{" + f"font-family: {font}; font-size: {font_size}px;" + "}"
    app.setStyleSheet(style_sheet)

    return app


# ==================================================
def SplineEx(points, n_points=None, closed=False, natural=True):
    spline_function = vtkParametricSpline()
    spline_function.SetPoints(pv.vtk_points(points, False))

    if closed:
        spline_function.ClosedOn()
    else:
        spline_function.ClosedOff()

    if natural and not closed:
        spline_function.SetLeftConstraint(2)
        spline_function.SetLeftValue(0.0)
        spline_function.SetRightConstraint(2)
        spline_function.SetRightValue(0.0)

    # get interpolation density
    u_res = n_points
    if u_res is None:
        u_res = points.shape[0]

    u_res -= 1
    spline = surface_from_para(spline_function, u_res)
    return spline.compute_arc_length()


# ==================================================
def create_unit_cell(A, origin, lower, dims):
    As = A.copy()

    # signle box.
    pts = np.array([[0.0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]])
    shift = np.array([origin] * len(pts))
    pts = pts + shift
    lines = np.array([5, 0, 1, 2, 3, 0, 5, 4, 5, 6, 7, 4, 2, 0, 4, 2, 1, 5, 2, 2, 6, 2, 3, 7])
    box = pv.PolyData(pts, lines=lines)
    box.transform(As)

    # repeated boxes.
    m = pv.UniformGrid(origin=lower, dims=dims).cast_to_unstructured_grid()
    m.transform(As)
    p = m.glyph(geom=box, factor=1.0)

    return box, p


# ==================================================
def axis_object(normA, labels, size, bold, italic, position):
    # axes arrows.
    a1, a2, a3 = normA
    mesh = pv.PolyData([[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0], [-1.1, -1.1, -1.1], [1.1, 1.1, 1.1]])

    colors = rcParams["detail.axis.color"] + [rcParams["detail.axis.color.center"]] + ["white", "white"]
    cmap = custom_colormap(colors)

    g0 = pv.Arrow(
        direction=a1,
        shaft_radius=rcParams["detail.axis.shaft.radius"],
        tip_radius=rcParams["detail.axis.tip.radius"],
        tip_length=rcParams["detail.axis.tip.length"],
        tip_resolution=rcParams["detail.axis.tip.resolution"],
    )
    g1 = pv.Arrow(
        direction=a2,
        shaft_radius=rcParams["detail.axis.shaft.radius"],
        tip_radius=rcParams["detail.axis.tip.radius"],
        tip_length=rcParams["detail.axis.tip.length"],
        tip_resolution=rcParams["detail.axis.tip.resolution"],
    )
    g2 = pv.Arrow(
        direction=a3,
        shaft_radius=rcParams["detail.axis.shaft.radius"],
        tip_radius=rcParams["detail.axis.tip.radius"],
        tip_length=rcParams["detail.axis.tip.length"],
        tip_resolution=rcParams["detail.axis.tip.resolution"],
    )
    g3 = pv.Sphere(
        rcParams["detail.axis.center.radius"],
        theta_resolution=rcParams["detail.axis.center.theta"],
        phi_resolution=rcParams["detail.axis.center.phi"],
    )
    g4 = pv.Sphere(0)
    idx = np.array([0, 1, 2, 3, 4, 5])
    mesh.point_data["scalars"] = idx
    axes = mesh.glyph(geom=(g0, g1, g2, g3, g4, g4), factor=rcParams["detail.axis.scale"], scale=False, rng=(0, 5))

    # axes labels.
    lbl_actor = vtk.vtkAxesActor()
    lbl_actor.SetShaftTypeToCylinder()
    lbl_actor.SetTipTypeToCone()
    lbl_actor.SetCylinderRadius(0.0)
    lbl_actor.SetConeRadius(0.0)
    lbl_actor.SetXAxisLabelText(labels[0])
    lbl_actor.SetYAxisLabelText(labels[1])
    lbl_actor.SetZAxisLabelText(labels[2])

    m = np.eye(4)
    m[0:3, 0] = a1
    m[0:3, 1] = a2
    m[0:3, 2] = a3
    transform = vtk.vtkTransform()
    transform.SetMatrix((m * position).ravel().tolist())
    lbl_actor.SetUserTransform(transform)

    x_p = lbl_actor.GetXAxisCaptionActor2D()
    y_p = lbl_actor.GetYAxisCaptionActor2D()
    z_p = lbl_actor.GetZAxisCaptionActor2D()
    for i, c, pos in zip([x_p, y_p, z_p], rcParams["detail.axis.label.color"], rcParams["detail.axis.label.position"]):
        i.GetPositionCoordinate().SetCoordinateSystemToWorld()
        i.GetPositionCoordinate().SetValue(*pos)
        i.GetTextActor().SetTextScaleModeToViewport()
        if bold:
            i.GetCaptionTextProperty().BoldOn()
        else:
            i.GetCaptionTextProperty().BoldOff()
        if italic:
            i.GetCaptionTextProperty().ItalicOn()
        else:
            i.GetCaptionTextProperty().ItalicOff()
        i.GetCaptionTextProperty().SetFontSize(size)
        rgb = all_colors[c][1]
        i.GetCaptionTextProperty().SetColor(*rgb)

    return axes, cmap, lbl_actor


# ==================================================
def plot_axis(plotter, normA, labels, size, bold, italic, position):
    axes, cmap, lbl_actor = axis_object(normA, labels, size, bold, italic, position)

    # axes.
    axes_actor = plotter.add_mesh(axes, show_scalar_bar=False, cmap=cmap, smooth_shading=True, name="axis")
    axes_widget = plotter.add_orientation_widget(axes_actor)
    axes_widget.SetViewport(*rcParams["detail.axis.viewport"])

    # axes labels.
    axes_label_widget = vtk.vtkOrientationMarkerWidget()
    axes_label_widget.SetOrientationMarker(lbl_actor)
    axes_label_widget.SetCurrentRenderer(plotter.renderer)
    axes_label_widget.SetInteractor(plotter.renderer.parent.iren.interactor)
    axes_label_widget.SetViewport(*rcParams["detail.axis.viewport"])
    axes_label_widget.SetEnabled(True)

    # set non-interactive for axes and labels.
    axes_widget.InteractiveOff()
    axes_label_widget.InteractiveOff()

    # axes actor.
    axes_actor = plotter.add_mesh(axes, show_scalar_bar=False, cmap=cmap, smooth_shading=True, name="axis")

    return axes_widget, axes_label_widget, axes_actor


# ==================================================
def view_vectors(A_norm, n):
    a1 = A_norm[0]
    a2 = A_norm[1]
    a3 = A_norm[2]
    n = NSArray(n, "vector", "value")
    if n.norm() < CHOP:
        return None, None

    v = NSArray(n[0] * a1 + n[1] * a2 + n[2] * a3, "vector", "value").normalize()

    if np.abs(v[2] - 1.0) < CHOP:  # v = (0,0,1)
        u = np.array([0, 1, 0])
    elif np.abs(v[2] + 1.0) < CHOP:  # v = (0,0,-1)
        u = np.array([0, -1, 0])
    else:
        vp = np.sqrt(1.0 - v[2] * v[2])
        u = np.array([-v[2] * v[0] / vp, -v[2] * v[1] / vp, vp])

    return v, u



"""
Qt Draw
"""
import os
from math import floor, ceil
import numpy as np
from qtpy.QtWidgets import QMessageBox, QFileDialog
from qtpy.uic import loadUi
import pyvista as pv
import pandas as pd
from gcoreutils.io_util import write_dict, read_dict
from gcoreutils.nsarray import NSArray
from gcoreutils.dataset import DataSet
from gcoreutils.crystal_util import cell_transform_matrix
from gcoreutils.latex_util import check_latex_installed, latex_cmd
from qtdraw.qt_draw_base import QtDrawBase
from qtdraw.core.setting import rcParams, preference
from qtdraw.core.util import create_unit_cell, plot_axis, view_vectors, create_application
from qtdraw.core.dialog_about import DialogAbout
from qtdraw.core.dialog_preference import DialogPreference
from qtdraw.core.line_edit import LineEditor
from qtdraw.core.group_tab import GroupTab
from qtdraw.core.basic_object import (
    create_sphere,
    create_bond,
    create_vector,
    create_orbital,
    create_stream_vector,
    create_plane,
    create_box,
    create_polygon,
    create_text,
    create_spline,
)
from qtdraw.core.color_palette import all_colors, custom_colormap, check_color
from qtdraw.parser.read_cif_vesta import plot_cif, plot_vesta
from qtdraw import __version__

from qtdraw.core.qt_logging import UncaughtHook

CHOP = 1e-5
qt_exception_hook = UncaughtHook()
global QT_EXCEPTION_HOOK
QT_EXCEPTION_HOOK = qt_exception_hook


# ==================================================
def to_chopped_list(a):
    a[np.abs(a) < CHOP] = 0.0
    return a.tolist()


# ==================================================
class QtDrawWidget(QtDrawBase):
    """
    QtDraw widget.
    """

    # ==================================================
    def __init__(
        self,
        title=None,
        model=None,
        cell=None,
        origin=None,
        view=None,
        size=None,
        axis_type=None,
        view_range=None,
        repeat=False,
        clip=True,
        cluster=False,
        background=False,
        parent=None,
    ):
        """
        initialize the class.

        Args:
            title (str, optional): window title of plotter.
            model (str, optioanl): name of model.
            cell (dict or str, optional): cell information (a/b/c/alpha/beta/gamma) or "hexagonal/trigonal".
            origin (list, optional): cell origin.
            view (list, optional): view with respect to the axes (a,b,c).
            size (list, optional): window size, (width, height).
            axis_type (str, optional): axis type, "xyz/abc/abc*".
            view_range (list, optional): display range ([x0,y0,z0],[x1,y1,z1]) (reduced).
            repeat (bool, optional): repeat plot ?
            clip (bool, optional): clip objects out of view_range ?
            cluster (bool, optional): cluster without repeat ?
            background (bool, optional): background run ?
            parent (QWidget, optional): parent.
        """
        self.preference = preference
        self.background = background
        # set qtdraw base.
        UI = os.path.dirname(__file__) + "/core/" + rcParams["plotter.ui"]
        UI = UI.replace(os.sep, "/")
        theme = rcParams["plotter.theme"]
        line = rcParams["detail.smooth.line"]
        point = rcParams["detail.smooth.point"]
        polygon = rcParams["detail.smooth.polygon"]
        shading = rcParams["detail.smooth.shading"]
        sample = rcParams["detail.smooth.samples"]
        super().__init__(
            parent=parent,
            panel=UI,
            theme=theme,
            smooth_line=line,
            smooth_point=point,
            smooth_polygon=polygon,
            smooth_shading=shading,
            sampling=sample,
        )

        # delete key event.
        for key in ["q", "b", "v", "C", "plus", "minus"]:
            self._layer.clear_events_for_key(key)

        # set anti aliasing.
        anti_aliasing = rcParams["detail.smooth.anti_aliasing"]
        self._set_anti_aliasing(anti_aliasing)

        # home folder.
        self._homedir = os.path.expanduser("~")
        os.chdir(self._homedir)

        # initialize.
        self._connect_slot()

        # set parameters.
        self._init_setting(title, model, cell, origin, view, size, axis_type, view_range, repeat, clip, cluster)

        self._create_panel()
        self._init_all()
        self.set_view()

    # ==================================================
    def _init_all(self):
        self._init_dataset()
        self._init_counter()
        self._init_panel()
        self._set_light()
        self._init_actor()

    # ==================================================
    def _init_setting(
        self,
        title=None,
        model=None,
        cell=None,
        origin=None,
        view=None,
        size=None,
        axis_type=None,
        view_range=None,
        repeat=False,
        clip=True,
        cluster=False,
    ):
        """
        Notes:
            - setting dict.
                - title: str
                - model: str
                - cell: (dict) a, b, c, alpha, beta, gamma
                - origin: [float, float, float]
                - view: [int, int, int]
                - size: [int, int]
                - axis.type: str
                - view_range: [ [float,float,float], [float,float,float] ]
                - repeat: bool
                - clip: bool
                - cluster: bool
        """
        if title is None:
            title = rcParams["init.title"]
        if model is None:
            model = rcParams["init.model"]
        if origin is None:
            origin = rcParams["init.origin"]
        if view is None:
            view = rcParams["init.view"]
        if size is None:
            size = rcParams["init.size"]
        if axis_type is None:
            axis_type = self.preference["axis.type"]
        if view_range is None:
            view_range = rcParams["init.view_range"]

        cell = cell_transform_matrix(cell)[0]
        lower = NSArray(view_range[0], "vector", "value").tolist()
        upper = NSArray(view_range[1], "vector", "value").tolist()

        # set Plotter setting.
        self.setting = {}
        self.setting["title"] = str(title)
        self.setting["model"] = str(model)
        self.setting["cell"] = cell
        self.setting["origin"] = NSArray(origin, "vector", "value").tolist()
        self.setting["view"] = NSArray(view, "vector", "value").astype(int).tolist()
        self.setting["size"] = NSArray(size, "vector", "value").astype(int).tolist()
        self.preference["axis.type"] = str(axis_type)
        self.setting["view_range"] = [lower, upper]
        self.setting["repeat"] = bool(repeat)
        self.setting["clip"] = bool(clip)
        self.setting["cluster"] = bool(cluster)
        self.setting["crystal"] = ""  # triclinic, monoclinic, orthorhombic, tetragonal, trigonal, hexagonal, cubic

        self.setting["axis_mode"] = 0
        self.setting["cell_mode"] = 2 if self.setting["cluster"] else 0

    # ==================================================
    def _create_panel(self):
        # model.
        self.edit_model_o = LineEditor(validator="string", callback=self._set_model, parent=self)
        self.layout_model.replaceWidget(self.edit_model, self.edit_model_o)

        # origin.
        self.edit_origin_o = LineEditor("Origin:", validator="r_vector", decimal=2, callback=self.set_origin, parent=self)
        self.layout_cell.replaceWidget(self.edit_origin, self.edit_origin_o)

        # a, b, c, alpha, beta, gamma.
        self.edit_a_o = LineEditor("a:", validator="real_positive", callback=self._set_a, parent=self)
        self.layout_cell.replaceWidget(self.edit_a, self.edit_a_o)
        self.edit_b_o = LineEditor("b:", validator="real_positive", callback=self._set_b, parent=self)
        self.layout_cell.replaceWidget(self.edit_b, self.edit_b_o)
        self.edit_c_o = LineEditor("c:", validator="real_positive", callback=self._set_c, parent=self)
        self.layout_cell.replaceWidget(self.edit_c, self.edit_c_o)
        self.edit_alpha_o = LineEditor(
            "α:",
            validator="real_positive",
            decimal=2,
            callback=self._set_alpha,
            parent=self,
        )
        self.layout_cell.replaceWidget(self.edit_alpha, self.edit_alpha_o)
        self.edit_beta_o = LineEditor(
            "β:",
            validator="real_positive",
            decimal=2,
            callback=self._set_beta,
            parent=self,
        )
        self.layout_cell.replaceWidget(self.edit_beta, self.edit_beta_o)
        self.edit_gamma_o = LineEditor(
            "γ:",
            validator="real_positive",
            decimal=2,
            callback=self._set_gamma,
            parent=self,
        )
        self.layout_cell.replaceWidget(self.edit_gamma, self.edit_gamma_o)

        # lower, upper.
        self.edit_lower_o = LineEditor(
            "lower:",
            validator="r_vector",
            decimal=2,
            callback=self._set_lower,
            parent=self,
        )
        self.layout_view.replaceWidget(self.edit_lower, self.edit_lower_o)
        self.edit_upper_o = LineEditor(
            "upper:",
            validator="r_vector",
            decimal=2,
            callback=self._set_upper,
            parent=self,
        )
        self.layout_view.replaceWidget(self.edit_upper, self.edit_upper_o)

        if self.setting["cluster"]:
            self.button_repeat.hide()

    # ==================================================
    def _set_light(self):
        if self.preference["light.eye_dome_lighting"]:
            self._layer.enable_eye_dome_lighting()
        else:
            self._layer.disable_eye_dome_lighting()
        self._layer.remove_all_lights()
        intensity = self.preference["light.intensity"]
        if self.preference["light.pbr"]:
            self._layer.add_light(pv.Light(light_type="headlight"))
            self._layer.add_light(pv.Light(light_type="headlight"))
        else:
            self._layer.add_light(pv.Light(light_type="headlight", intensity=0.55))
        self._layer.add_light(pv.Light(light_type="headlight", intensity=intensity))

    # ==================================================
    def _init_panel(self):
        # window size.
        self.set_window_size(*self.setting["size"])

        # axis object.
        self.button_object.setChecked(False)

        # set parallel projection.
        projection = rcParams["init.parallel_projection"]
        self.button_parallel.setChecked(projection)
        self._toggle_parallel_projection(projection)

        # repeat status.
        self.button_repeat.setChecked(self.setting["repeat"])

        # clip status.
        self.button_clip.setChecked(self.setting["clip"])

        # set grid.
        self._grid_actor = self._layer.show_grid()
        grid_status = rcParams["init.grid"]
        self.button_grid.setChecked(grid_status)
        self._toggle_grid(grid_status)

        self._set_model(self.setting["model"])
        self.set_range(self.setting["view_range"])
        self.set_origin(self.setting["origin"])
        self.set_crystal(self.setting["crystal"])

        self._change_axis(self.setting["axis_mode"])
        self._change_cell(self.setting["cell_mode"])

    # ==================================================
    def _load_plus_panel(self, fname, current=None):
        if current is None:
            current = ""
        else:
            current = current.replace(os.sep, "/")
            current = current[: current.rfind("/")]

        loadUi(current + "/" + fname, self.plus_panel)
        self._set_plus_panel()

    # ==================================================
    def set_crystal(self, crystal):
        """
        set crystal.

        Args:
            crystal (str): crystal, "triclinic/monoclinic/orthorhombic/tetragonal/trigonal/hexagonal/cubic".
        """
        if crystal == "monoclinic":
            ro = [False, False, False, True, False, True]
        elif crystal == "orthorhombic":
            ro = [False, False, False, True, True, True]
        elif crystal in ["tetragonal", "trigonal", "hexagonal"]:
            ro = [False, True, False, True, True, True]
        elif crystal == "cubic":
            ro = [False, True, True, True, True, True]
        elif crystal in ["", "triclinic"]:
            ro = [False, False, False, False, False, False]
            crystal = ""
        else:
            raise KeyError(f"{crystal} is invalid.")

        self.edit_a_o.editor.setReadOnly(ro[0])
        self.edit_b_o.editor.setReadOnly(ro[1])
        self.edit_c_o.editor.setReadOnly(ro[2])
        self.edit_alpha_o.editor.setReadOnly(ro[3])
        self.edit_beta_o.editor.setReadOnly(ro[4])
        self.edit_gamma_o.editor.setReadOnly(ro[5])
        self.setting["crystal"] = crystal
        self._set_axis_cell(self.setting["cell"])

    # ==================================================
    def set_range(self, view_range=None):
        """
        set display range.

        Args:
            view_range (list, optional): display range, [ [x0,y0,z0], [x1,y1,z1] ] (reduced).

        Notes:
            - if view_range is None, default, [[0,0,0],[1,1,1]] is used.
        """
        if view_range is None:
            view_range = ["[0.0,0.0,0.0]", "[1.0,1.0,1.0]"]

        if type(view_range) is not list:
            raise KeyError(f"{type(view_range)} is invalid.")

        lower, upper = view_range
        lower = NSArray(lower, "vector", "value").tolist()
        upper = NSArray(upper, "vector", "value").tolist()
        self.setting["view_range"] = [lower, upper]
        self.edit_lower_o.setText(str(lower))
        self.edit_upper_o.setText(str(upper))

        i1 = [floor(lower[0]), floor(lower[1]), floor(lower[2])]
        i2 = [ceil(upper[0]), ceil(upper[1]), ceil(upper[2])]
        dims = [i2[0] - i1[0], i2[1] - i1[1], i2[2] - i1[2]]

        self._ilower = i1
        self._dims = dims

        if dims[0] * dims[1] * dims[2] == 0:
            self._igrid = NSArray("[0,0,0]", fmt="value").astype(int).tolist()
        else:
            self._igrid = NSArray.igrid(dims, i1).value().astype(int).tolist()

        self._set_axis_cell(self.setting["cell"])

    # ==================================================
    def set_origin(self, origin=None):
        """
        set origin.

        Args:
            origin (list, optional): origin (reduced).

        Notes:
            - if origin is None, default, [0,0,0] is used.
        """
        if origin is None:
            origin = "[0,0,0]"

        origin = NSArray(origin, "vector", "value").tolist()
        self.setting["origin"] = origin
        self._set_axis_cell(self.setting["cell"])
        self.edit_origin_o.setText(str(origin))

    # ==================================================
    def set_view(self, view=None):
        """
        set view point.

        Args:
            view (list, optional): view point with respect to the axes (a,b,c)

        Notes:
            - if view is None, current view is used.
        """
        if view is None:
            view = self.setting["view"]
        else:
            view = NSArray(view, "vector", "value").astype(int).tolist()

        self.setting["view"] = view

        v, u = view_vectors(self._A_norm, view)
        if v is not None:
            a, b, c = view
            self.combo_a.setCurrentIndex(a + 9)
            self.combo_b.setCurrentIndex(b + 9)
            self.combo_c.setCurrentIndex(c + 9)
            self._layer.view_vector(v, u)

    # ==================================================
    def set_axis(self, mode):
        """
        set axis mode.

        Args:
            mode (str): mode, "on/only/off".
        """
        mode_idx = {"on": 0, "only": 1, "off": 2}

        if mode not in mode_idx.keys():
            return

        self._change_axis(mode_idx[mode])

    # ==================================================
    def set_cell(self, mode):
        """
        set cell mode.

        Args:
            mode (str): mode, "single/all/off".
        """
        mode_idx = {"single": 0, "all": 1, "off": 2}

        if mode not in mode_idx.keys():
            return

        self._change_cell(mode_idx[mode])

    # ==================================================
    def _regularize_position(self, position):
        if not isinstance(position, (str, list, np.ndarray)):
            raise KeyError(f"{type(position)} is invalid.")

        position = NSArray(position, "vector", "value")
        if position.lst:
            position = position.tolist()
        else:
            position = [position.tolist()]

        return position

    # ==================================================
    def plot_site(
        self, position, size=None, color=None, opacity=None, space=None, name=None, label="", show_lbl=False, cell=[0, 0, 0]
    ):
        """
        plot site.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot site (reduced).
            size (float, optional): size of site.
            color (str, optional): color of site.
            opacity (float, optional): opacity of site.
            space (int, optional): pre-space of label.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - if argument is None, default value is used.
        """
        position = self._regularize_position(position)

        if size is None:
            size = self.def_value["site"][5]
        if color is None:
            color = self.def_value["site"][6]
        if opacity is None:
            opacity = self.def_value["site"][7]
        if space is None:
            space = self.def_value["site"][8]

        if size < CHOP:
            return
        if space < 0:
            return

        info = [float(size), color, float(opacity), int(space)]
        self._plot_to_data("site", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_bond(
        self,
        position,
        vector=None,
        width=None,
        color=None,
        color2="",
        opacity=None,
        space=None,
        name=None,
        label="",
        show_lbl=False,
        cell=[0, 0, 0],
    ):
        """
        plot bond.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot bond center (reduced).
            vector (str or list or ndarray or NSArray, optional): bond vector (reduced).
            width (float, optional): width of bond.
            color (str, optional): color of bond.
            color2 (str, optional): color of bond for half tail.
            opacity (float, optional): opacity of bond.
            space (int, optional): pre-space of label.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - if argument is None, default value is used.
            - if color2 is "", same color is used.
        """
        position = self._regularize_position(position)

        if vector is None:
            vector = self.def_value["bond"][5]
        if width is None:
            width = self.def_value["bond"][6]
        if color is None:
            color = self.def_value["bond"][7]
        if color2 is None:
            color2 = self.def_value["bond"][8]
        if opacity is None:
            opacity = self.def_value["bond"][9]
        if space is None:
            space = self.def_value["bond"][10]

        if space < 0:
            return

        if color2 == "":
            color2 = color

        v = NSArray(vector, "vector", "value")
        if v.norm() < CHOP or width < CHOP:
            return

        info = [to_chopped_list(v), float(width), color, color2, float(opacity), int(space)]
        self._plot_to_data("bond", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_vector(
        self,
        position,
        vector=None,
        length=None,
        width=None,
        offset=None,
        color=None,
        opacity=None,
        space=None,
        name=None,
        label="",
        show_lbl=False,
        cell=[0, 0, 0],
    ):
        """
        plot vector.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot vector (reduced).
            vector (str or list or ndarray or NSArray, optional): vector direction (cartesian).
            length (float, optional): length of vector.
            width (float, optional): width of vector.
            offset (float, optional): offset of vector end (ratio).
            color (str, optional): color of vector.
            opacity (float, optional): opacity of vector.
            space (int, optional): pre-space of label.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - if argument is None, default value is used.
            - vector is not necessary to be normalized.
        """
        position = self._regularize_position(position)

        if vector is None:
            vector = self.def_value["vector"][5]
        if length is None:
            length = self.def_value["vector"][6]
        if width is None:
            width = self.def_value["vector"][7]
        if offset is None:
            offset = self.def_value["vector"][8]
        if color is None:
            color = self.def_value["vector"][9]
        if opacity is None:
            opacity = self.def_value["vector"][10]
        if space is None:
            space = self.def_value["vector"][11]

        if space < 0:
            return

        v = NSArray(vector, "vector", "value")
        if v.norm() < CHOP or length < CHOP or width < CHOP:
            return

        info = [to_chopped_list(v), float(length), float(width), float(offset), color, float(opacity), int(space)]
        self._plot_to_data("vector", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_orbital(
        self,
        position,
        shape=None,
        surface="",
        size=None,
        theta_range=None,
        phi_range=None,
        color=None,
        opacity=None,
        space=None,
        scale=True,
        name=None,
        label="",
        show_lbl=False,
        cell=[0, 0, 0],
    ):
        """
        plot orbital.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot orbital (reduced).
            shape (str, optional): (x,y,z) shape of orbital in terms of (x,y,z) (cartesian).
            surface (str, optional): surface color in terms of (x,y,z) (cartesian).
            size (float, optional): size of orbital.
            theta_range (str or list or ndarray or NSArray, optional): theta range, default=[0,180].
            phi_range (str or list or ndarray or NSArray, optional): phi range, default=[0,360].
            color (str, optional): color/colormap of surface.
            opacity (float, optional): opacity of orbital.
            space (int, optional): pre-space of label.
            scale (bool, optional): if False, absolute size is used.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - if surface is "", surface is same as shape.
            - if argument is None, default value is used.
        """
        position = self._regularize_position(position)

        if shape is None:
            shape = self.def_value["orbital"][5]
        if surface is None:
            surface = self.def_value["orbital"][6]
        if size is None:
            size = self.def_value["orbital"][7]
        if scale is None:
            scale = self.def_value["orbital"][8]
        if theta_range is None:
            theta_range = [self.def_value["orbital"][9], self.def_value["orbital"][10]]
        if phi_range is None:
            phi_range = [self.def_value["orbital"][11], self.def_value["orbital"][12]]
        if color is None:
            color = self.def_value["orbital"][13]
        if opacity is None:
            opacity = self.def_value["orbital"][14]
        if space is None:
            space = self.def_value["orbital"][15]

        if size < CHOP:
            return

        if space < 0:
            return

        if type(shape) is not str:
            shape = str(shape)
        if surface == "":
            surface = shape
        if type(surface) is not str:
            surface = str(surface)

        theta_range = NSArray(theta_range, "vector", "value").tolist()
        phi_range = NSArray(phi_range, "vector", "value").tolist()

        info = [
            shape,
            surface,
            float(size),
            bool(scale),
            int(theta_range[0]),
            int(theta_range[1]),
            int(phi_range[0]),
            int(phi_range[1]),
            color,
            float(opacity),
            int(space),
        ]
        self._plot_to_data("orbital", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_stream_vector(
        self,
        position,
        shape=None,
        vector=None,
        size=None,
        v_size=None,
        width=None,
        scale=None,
        theta=None,
        phi=None,
        theta_range=None,
        phi_range=None,
        color=None,
        component=None,
        opacity=None,
        space=None,
        name=None,
        label="",
        show_lbl=False,
        cell=[0, 0, 0],
    ):
        """
        plot stream vector with center orbital.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot stream vector (reduced).
            shape (str, optional): shape of center orbital in terms of (x,y,z) (cartesian).
            vector (str or list or ndarray or NSArray, optional): vector components in terms of (x,y,z) [str]/str (cartesian).
            size (float, optional): size of orbital.
            v_size (float, optional): ratio of vector size to orbital.
            width (float, optional): width of vector.
            scale (bool, optional): scale vector size by value ?
            theta (int, optional): resolution in theta direction.
            phi (int, optional): resolution in phi direction.
            theta_range (str or list or ndarray or NSArray, optional): theta range, default=[0,180].
            phi_range (str or list or ndarray or NSArray, optional): phi range, default=[0,360].
            color (str, optional): color of vector.
            component (int, optional): component of vector to use for colormap (0,1,2,None = x,y,z,abs.).
            opacity (float, optional): opacity of center orbital.
            space (int, optional): pre-space of label.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - if shape is None, "1" is used.
            - if argument is None, default value is used.
        """
        position = self._regularize_position(position)

        if shape is None:
            shape = self.def_value["stream"][5]
        if vector is None:
            vector = self.def_value["stream"][6]
        if size is None:
            size = self.def_value["stream"][7]
        if v_size is None:
            v_size = self.def_value["stream"][8]
        if width is None:
            width = self.def_value["stream"][9]
        if scale is None:
            scale = self.def_value["stream"][10]
        if theta is None:
            theta = self.def_value["stream"][11]
        if phi is None:
            phi = self.def_value["stream"][12]
        if theta_range is None:
            theta_range = [self.def_value["stream"][13], self.def_value["stream"][14]]
        if phi_range is None:
            phi_range = [self.def_value["stream"][15], self.def_value["stream"][16]]
        if color is None:
            color = self.def_value["stream"][17]
        if component is None:
            component = self.def_value["stream"][18]
        if opacity is None:
            opacity = self.def_value["stream"][19]
        if space is None:
            space = self.def_value["stream"][20]

        if space < 0:
            return

        if type(shape) is not str:
            shape = str(shape)

        vector = NSArray(vector, "vector").str()
        theta_range = NSArray(theta_range, "vector", "value").tolist()
        phi_range = NSArray(phi_range, "vector", "value").tolist()

        info = [
            shape,
            vector,
            float(size),
            float(v_size),
            float(width),
            bool(scale),
            int(theta),
            int(phi),
            int(theta_range[0]),
            int(theta_range[1]),
            int(phi_range[0]),
            int(phi_range[1]),
            color,
            component,
            float(opacity),
            int(space),
        ]
        self._plot_to_data("stream", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_plane(
        self,
        position,
        normal=None,
        x=None,
        y=None,
        color=None,
        opacity=None,
        space=None,
        name=None,
        label="",
        show_lbl=False,
        cell=[0, 0, 0],
    ):
        """
        plot plane.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot plane (reduced).
            normal (str or list or ndarray or NSArray, optional): normal vector of plane (reduced).
            x (float, optional): x size of plane.
            y (float, optional); y size of plane.
            color (str, optional): color of plane.
            opacity (float, optional): opacity of plane.
            space (int, optional): pre-space of label.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - position is for center of plane.
            - if argument is None, default value is used.
            - normal is not necessary to be normalized.
        """
        position = self._regularize_position(position)

        if normal is None:
            normal = self.def_value["plane"][5]
        if x is None:
            x = self.def_value["plane"][6]
        if y is None:
            y = self.def_value["plane"][7]
        if color is None:
            color = self.def_value["plane"][8]
        if opacity is None:
            opacity = self.def_value["plane"][9]
        if space is None:
            space = self.def_value["plane"][10]

        if space < 0:
            return

        v = NSArray(normal, "vector", "value")
        if v.norm() < CHOP or x < CHOP or y < CHOP:
            return

        info = [to_chopped_list(v), float(x), float(y), color, float(opacity), int(space)]
        self._plot_to_data("plane", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_box(
        self,
        position,
        a1=None,
        a2=None,
        a3=None,
        edge=None,
        wireframe=None,
        width=None,
        color=None,
        opacity=None,
        space=None,
        name=None,
        label="",
        show_lbl=False,
        cell=[0, 0, 0],
    ):
        """
        plot box.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot box (reduced).
            a1 (str or list or ndarray or NSArray, optional): 1st direction of box (reduced).
            a2 (str or list or ndarray or NSArray, optional): 2nd direction of box (reduced).
            a3 (str or list or ndarray or NSArray, optional): 3rd direction of box (reduced).
            edge (bool, optional): show edge of box ?
            wireframe (bool, optional): draw box by wireframe ?
            width (float, optional): line width of box for wireframe plot.
            color (str, optional): color of box.
            opacity (float, optional): opacity of box.
            space (int, optional): pre-space of label.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - if argument is None, default value is used.
        """
        position = self._regularize_position(position)

        if a1 is None:
            a1 = self.def_value["box"][5]
        if a2 is None:
            a2 = self.def_value["box"][6]
        if a3 is None:
            a3 = self.def_value["box"][7]
        if edge is None:
            edge = self.def_value["box"][8]
        if wireframe is None:
            wireframe = self.def_value["box"][9]
        if width is None:
            width = self.def_value["box"][10]
        if color is None:
            color = self.def_value["box"][11]
        if opacity is None:
            opacity = self.def_value["box"][12]
        if space is None:
            space = self.def_value["box"][13]

        if space < 0:
            return

        a1 = NSArray(a1, "vector", "value")
        a2 = NSArray(a2, "vector", "value")
        a3 = NSArray(a3, "vector", "value")
        if a1.norm() < CHOP or a2.norm() < CHOP or a3.norm() < CHOP or width < CHOP:
            return

        info = [
            to_chopped_list(a1),
            to_chopped_list(a2),
            to_chopped_list(a3),
            bool(edge),
            bool(wireframe),
            float(width),
            color,
            float(opacity),
            int(space),
        ]
        self._plot_to_data("box", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_polygon(
        self,
        position,
        point=None,
        connection=None,
        edge=None,
        wireframe=None,
        width=None,
        color=None,
        opacity=None,
        space=None,
        name=None,
        label="",
        show_lbl=False,
        cell=[0, 0, 0],
    ):
        """
        plot polygon.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot polygon (reduced).
            point (str or list or ndarray or NSArray, optional): vertices of polygon (reduced).
            connection (str or list or ndarray or NSArray, optional): list of connected vectices for each plane.
            edge (bool, optional): show edge of box ?
            wireframe (bool, optional): draw box by wireframe ?
            width (float, optional): line width of box for wireframe plot.
            color (str, optional): color of box.
            opacity (float, optional): opacity of box.
            space (int, optional): pre-space of label.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - if argument is None, default value is used.
        """
        position = self._regularize_position(position)

        if point is None:
            point = self.def_value["polygon"][5]
        if connection is None:
            connection = self.def_value["polygon"][6]
        if edge is None:
            edge = self.def_value["polygon"][7]
        if wireframe is None:
            wireframe = self.def_value["polygon"][8]
        if width is None:
            width = self.def_value["polygon"][9]
        if color is None:
            color = self.def_value["polygon"][10]
        if opacity is None:
            opacity = self.def_value["polygon"][11]
        if space is None:
            space = self.def_value["polygon"][12]

        if space < 0:
            return

        point = to_chopped_list(NSArray(point, "vector", "value"))
        connection = NSArray(connection, "vector", "value").astype(int).tolist()

        info = [point, connection, bool(edge), bool(wireframe), float(width), color, float(opacity), int(space)]
        self._plot_to_data("polygon", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_text3d(
        self,
        position,
        text=None,
        size=None,
        depth=None,
        normal=None,
        offset=None,
        color=None,
        opacity=None,
        space=None,
        name=None,
        label="",
        show_lbl=False,
        cell=[0, 0, 0],
    ):
        """
        plot 3d text.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot 3d text (reduced).
            text (str, optional): text.
            size (float, optional): size of text.
            depth (float, optional): depth of text.
            normal (str or list or ndarray or NSArray, optional): normal vector of text (reduced).
            offset (str or list or ndarray or NSArray, optional): offset of text.
            color (str, optional): color of text.
            opacity (float, optional): opacity of text.
            space (int, optional): pre-space of label.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - if argument is None, default value is used.
            - normal is not necessary to be normalized.
        """
        position = self._regularize_position(position)

        if text is None:
            text = self.def_value["text3d"][5]
        if size is None:
            size = self.def_value["text3d"][6]
        if depth is None:
            depth = self.def_value["text3d"][7]
        if normal is None:
            normal = self.def_value["text3d"][8]
        if offset is None:
            offset = self.def_value["text3d"][9]
        if color is None:
            color = self.def_value["text3d"][10]
        if opacity is None:
            opacity = self.def_value["text3d"][11]
        if space is None:
            space = self.def_value["text3d"][12]

        if space < 0:
            return

        if text == "" or size < CHOP:
            return
        n = NSArray(normal, "vector", "value")
        if n.norm() < CHOP:
            return

        offset = to_chopped_list(NSArray(offset, "vector", "value"))

        info = [text, float(size), float(depth), to_chopped_list(n), offset, color, float(opacity), int(space)]
        self._plot_to_data("text3d", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_spline(
        self,
        position,
        point=None,
        width=None,
        n_interp=None,
        closed=None,
        natural=None,
        color=None,
        opacity=None,
        space=None,
        name=None,
        label="",
        show_lbl=False,
        cell=[0, 0, 0],
    ):
        """
        plot spline curve.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot spline curve (reduced).
            point (str or list or ndarray or NSArray, optional): points to be interpolated (reduced).
            width (float, optioanl): width of spline curve.
            n_interp (int, optional): number of interpolation points.
            closed (bool, optional): closed spline ?
            natural (bool, optional): natural boundary ?
            color (str, optional): color of spline curve.
            opacity (float, optional): opacity of spline curve.
            space (int, optional): pre-space of label.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - if argument is None, default value is used.
        """
        position = self._regularize_position(position)

        if point is None:
            point = self.def_value["spline"][5]
        if width is None:
            width = self.def_value["spline"][6]
        if n_interp is None:
            n_interp = self.def_value["spline"][7]
        if closed is None:
            closed = self.def_value["spline"][8]
        if natural is None:
            natural = self.def_value["spline"][9]
        if color is None:
            color = self.def_value["spline"][10]
        if opacity is None:
            opacity = self.def_value["spline"][11]
        if space is None:
            space = self.def_value["spline"][12]

        if width < CHOP:
            return

        if space < 0:
            return

        point = to_chopped_list(NSArray(point, "vector", "value"))

        info = [point, float(width), int(n_interp), bool(closed), bool(natural), color, float(opacity), int(space)]
        self._plot_to_data("spline", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_spline_t(
        self,
        position,
        expression=None,
        t_range=None,
        width=None,
        n_interp=None,
        closed=None,
        natural=None,
        color=None,
        opacity=None,
        space=None,
        name=None,
        label="",
        show_lbl=False,
        cell=[0, 0, 0],
    ):
        """
        plot spline curve.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot spline curve (reduced).
            expression (str or list or ndarray or NSArray, optional): component function of "t" to create interpolated points (reduced).
            t_range (str or list or ndarray or NSArray, optional): range of "t", [start, stop, step].
            width (float, optioanl): width of spline curve.
            n_interp (int, optional): number of interpolation points.
            closed (bool, optional): closed spline ?
            natural (bool, optional): natural boundary ?
            color (str, optional): color of spline curve.
            opacity (float, optional): opacity of spline curve.
            space (int, optional): pre-space of label.
            name (str, optional): group name of object.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.

        Notes:
            - if argument is None, default value is used.
        """
        position = self._regularize_position(position)

        if expression is None:
            expression = self.def_value["spline_t"][5]
        if t_range is None:
            t_range = self.def_value["spline_t"][6]
        if width is None:
            width = self.def_value["spline_t"][7]
        if n_interp is None:
            n_interp = self.def_value["spline_t"][8]
        if closed is None:
            closed = self.def_value["spline_t"][9]
        if natural is None:
            natural = self.def_value["spline_t"][10]
        if color is None:
            color = self.def_value["spline_t"][11]
        if opacity is None:
            opacity = self.def_value["spline_t"][12]
        if space is None:
            space = self.def_value["spline_t"][13]

        if width < CHOP:
            return

        if space < 0:
            return

        expression = NSArray(expression, "vector").str()
        t_range = to_chopped_list(NSArray(t_range, "vector", "value"))

        info = [expression, t_range, float(width), int(n_interp), bool(closed), bool(natural), color, float(opacity), int(space)]
        self._plot_to_data("spline_t", name, position, info, label, show_lbl, cell)

    # ==================================================
    def plot_caption(self, position, caption=None, space=None, size=None, color=None, bold=None, name=None, cell=[0, 0, 0]):
        """
        plot caption.

        Args:
            position (str or list or ndarray or NSArray): position(s) to plot labels (reduced).
            caption (str or list, optional): label or list of labels.
            space (int, optional): pre-space of label.
            size (int, optional): font size.
            color (str, optional): text color.
            bold (bool, optional): bold face ?
            name (str, optional): group name.
            cell (list, optional): cell position.

        Notes:
            - if caption is None, simple number is used.
            - if argument is None, default value is used.
            - list size of caption must be equal to that of position.
        """
        position = self._regularize_position(position)

        if caption is None:
            caption = [str(i) for i in range(len(position))]
        if space is None:
            space = self.def_value["caption"][4]
        if size is None:
            size = self.def_value["caption"][5]
        if color is None:
            color = self.def_value["caption"][6]
        if bold is None:
            bold = self.def_value["caption"][7]

        if space < 0:
            return

        if type(caption) == str:
            caption = [caption] * len(position)

        if len(position) != len(caption):
            return

        info = [caption, int(space), int(size), bool(bold), color]
        self._plot_to_data("caption", name, position, info, cell=cell)

    # ==================================================
    def plot_text(self, position, caption=None, relative=None, size=None, color=None, font=None, name=None):
        """
        plot text.

        Args:
            position (str or list or ndarray or NSArray): position to plot caption, [x,y].
            caption (str, optional): label.
            relative (bool, optional): relative position ?
            size (int, optional): font size.
            color (str, optional): text color.
            font (str, optional): font, "arial/times/courier".
            name (str, optional): group name.

        Notes:
            - if argument is None, default value is used.
        """
        position = self._regularize_position(position)[0]

        if relative is None:
            relative = self.def_value["text"][2]
        if caption is None:
            caption = self.def_value["text"][3]
        if size is None:
            size = self.def_value["text"][4]
        if color is None:
            color = self.def_value["text"][5]
        if font is None:
            font = self.def_value["text"][6]

        info = [relative, caption, int(size), color, font]
        self._plot_to_data("text", name, position, info)

    # ==================================================
    def show(self):
        """
        show the plot.

        Notes:
            - this must be called at the end of plots.
        """
        self._plot_all_object()
        if not self.background:
            super().show()

    # ==================================================
    def _connect_slot(self):
        self.button_x.clicked.connect(self._view_x)
        self.button_y.clicked.connect(self._view_y)
        self.button_z.clicked.connect(self._view_z)
        self.button_default.clicked.connect(self._view_d)
        self.button_screenshot.clicked.connect(lambda _: self._screenshot())
        self.button_list.clicked.connect(self._view_dataset)
        self.button_clear.clicked.connect(self._clear)
        self.button_load.clicked.connect(self._load)
        self.button_save.clicked.connect(self._save)

        self.button_repeat.toggled["bool"].connect(self._toggle_repeat)
        self.button_clip.toggled["bool"].connect(self._toggle_clip)
        self.button_parallel.toggled["bool"].connect(self._toggle_parallel_projection)
        self.button_grid.toggled["bool"].connect(self._toggle_grid)
        self.button_object.toggled["bool"].connect(self._toggle_axis_object)

        self.combo_a.currentIndexChanged["int"].connect(self._change_a)
        self.combo_b.currentIndexChanged["int"].connect(self._change_b)
        self.combo_c.currentIndexChanged["int"].connect(self._change_c)
        self.combo_axis.currentIndexChanged["int"].connect(self._change_axis)
        self.combo_cell.currentIndexChanged["int"].connect(self._change_cell)

        self.button_preference.clicked.connect(self._set_preference)
        self.button_about.clicked.connect(self._about)

        try:
            from multipie import __version__

            self.button_multipie.clicked.connect(self._multipie)
            self._multipie_loaded = __version__
        except ImportError:
            self._multipie_loaded = False
            self.button_multipie.hide()

    # ==================================================
    def _init_dataset(self):
        # dataset panel status.
        self._dialog_dataset = False

        # dataset.
        self.dataset = DataSet(rcParams["detail.dataset.property"])

        # default value.
        self.def_value = {
            group: [self.dataset.role[group][i][-1] for i in range(len(self.dataset.role[group]))]
            for group in self.dataset.keys()
        }

    # ==================================================
    def _init_counter(self):
        # init counter.
        self._counter = {i: 0 for i in self.dataset.keys()}

    # ==================================================
    def _init_actor(self):
        self._actorset = {}
        self._counter["actor"] = 0
        self._spotlight_actor = []

    # ==================================================
    def _set_model(self, model):
        self.setting["model"] = model
        title = self.setting["title"]
        self.set_window_title(f"{title} - {model}")
        self.edit_model_o.setText(model)

    # ==================================================
    def _set_axis_cell(self, cell=None):
        """
        Notes:
            - assume origin and range (ilower and dims) are already set.
        """
        if cell is None:
            cell = {}

        cell, self._volume, self._A, self._G, self._A_norm = cell_transform_matrix(cell, crystal=self.setting["crystal"])
        origin = self.setting["origin"]

        self.setting["cell"] = cell

        # set axis.
        self._draw_axis()

        # set volume.
        self.label_volume.setText(f"Volume: {self._volume:.6f}")

        # set cell.
        self.uc = create_unit_cell(self._A, origin, self._ilower, self._dims)
        self._draw_cell()

        self.edit_a_o.setText(str(self.setting["cell"]["a"]))
        self.edit_b_o.setText(str(self.setting["cell"]["b"]))
        self.edit_c_o.setText(str(self.setting["cell"]["c"]))
        self.edit_alpha_o.setText(str(self.setting["cell"]["alpha"]))
        self.edit_beta_o.setText(str(self.setting["cell"]["beta"]))
        self.edit_gamma_o.setText(str(self.setting["cell"]["gamma"]))

    # ==================================================
    def _draw_axis(self, mode=None):
        if mode is None:
            mode = self.combo_axis.currentIndex()

        self._screen_off()
        self._axis_actor = plot_axis(
            self._layer,
            self._A_norm,
            rcParams["detail.axis.label"][self.preference["axis.type"]],
            size=self.preference["axis.size"],
            bold=self.preference["axis.bold"],
            italic=self.preference["axis.italic"],
            position=self.preference["axis.position"],
        )
        self._axis_actor[2].SetVisibility(False)
        self._change_axis(mode)
        self._screen_on()

    # ==================================================
    def _draw_cell(self, mode=None):
        if mode is None:
            mode = self.combo_cell.currentIndex()

        self._screen_off()
        cell_actor = self._layer.add_mesh(
            self.uc[0],
            line_width=self.preference["cell.width"],
            color=all_colors[self.preference["cell.color"]][0],  # hex
            opacity=self.preference["cell.opacity"],
            smooth_shading=self.shading,
            name="uc",
        )

        if self.uc[1].n_points > 0:
            cell_actor1 = self._layer.add_mesh(
                self.uc[1],
                line_width=self.preference["cell.width"],
                color=all_colors[self.preference["cell.color"]][0],  # hex
                opacity=self.preference["cell.opacity"],
                smooth_shading=self.shading,
                name="uc1",
            )
        else:
            cell_actor1 = None

        self._cell_actor = [cell_actor, cell_actor1]
        self._change_cell(mode)
        self._screen_on()

    # ==================================================
    def _set_lower(self, lower):
        self.set_range([lower, self.setting["view_range"][1]])
        self._set_repeat()

    # ==================================================
    def _set_upper(self, upper):
        self.set_range([self.setting["view_range"][0], upper])
        self._set_repeat()

    # ==================================================
    def _set_a(self, a):
        self.setting["cell"]["a"] = float(a)
        self._set_axis_cell(self.setting["cell"])
        self._plot_all_object()

    # ==================================================
    def _set_b(self, b):
        self.setting["cell"]["b"] = float(b)
        self._set_axis_cell(self.setting["cell"])
        self._plot_all_object()

    # ==================================================
    def _set_c(self, c):
        self.setting["cell"]["c"] = float(c)
        self._set_axis_cell(self.setting["cell"])
        self._plot_all_object()

    # ==================================================
    def _set_alpha(self, alpha):
        self.setting["cell"]["alpha"] = float(alpha)
        self._set_axis_cell(self.setting["cell"])
        self._plot_all_object()

    # ==================================================
    def _set_beta(self, beta):
        self.setting["cell"]["beta"] = float(beta)
        self._set_axis_cell(self.setting["cell"])
        self._plot_all_object()

    # ==================================================
    def _set_gamma(self, gamma):
        self.setting["cell"]["gamma"] = float(gamma)
        self._set_axis_cell(self.setting["cell"])
        self._plot_all_object()

    # ==================================================
    def _toggle_repeat(self, flag):
        self.setting["repeat"] = flag
        self._close_dialog()
        self._repeat_data()
        if self.setting["repeat"]:
            self._change_cell(1)
        else:
            self._change_cell(0)

    # ==================================================
    def _toggle_clip(self, flag):
        self.setting["clip"] = flag
        self._plot_all_object()

    # ==================================================
    def _view_x(self):
        self.set_view([1, 0, 0])

    # ==================================================
    def _view_y(self):
        self.set_view([0, 1, 0])

    # ==================================================
    def _view_z(self):
        self.set_view([0, 0, 1])

    # ==================================================
    def _view_d(self):
        self.set_view(rcParams["init.view"])

    # ==================================================
    def _toggle_grid(self, flag):
        self._grid_actor.SetVisibility(flag)

    # ==================================================
    def _change_a(self, val):
        _, b, c = self.setting["view"]
        self.set_view([val - 9, b, c])

    # ==================================================
    def _change_b(self, val):
        a, _, c = self.setting["view"]
        self.set_view([a, val - 9, c])

    # ==================================================
    def _change_c(self, val):
        a, b, _ = self.setting["view"]
        self.set_view([a, b, val - 9])

    # ==================================================
    def _toggle_parallel_projection(self, flag):
        self._set_parallel_projection(flag)

    # ==================================================
    def _change_axis(self, val):
        ax = {0: (True, True), 1: (True, False), 2: (False, False)}

        show = ax[val]
        self._axis_actor[0].SetEnabled(show[0])
        self._axis_actor[1].SetEnabled(show[1])
        self.combo_axis.setCurrentIndex(val)

    # ==================================================
    def _toggle_axis_object(self, flag):
        self._axis_actor[2].SetVisibility(flag)
        if flag:
            self.axis_vis = self.combo_axis.currentIndex()
            self.cell_vis = self.combo_cell.currentIndex()
            self._hide_all_object()
            self._change_axis(2)
            self._change_cell(2)
        else:
            self._change_axis(self.axis_vis)
            self._change_cell(self.cell_vis)
            self._show_all_object()

    # ==================================================
    def _change_cell(self, val):
        uc = {0: (True, False), 1: (False, True), 2: (False, False)}
        self._cell_actor[0].SetVisibility(uc[val][0])
        if self._cell_actor[1] is not None:
            self._cell_actor[1].SetVisibility(uc[val][1])
        self.combo_cell.setCurrentIndex(val)

    # ==================================================
    def _set_repeat(self, repeat=None):
        if repeat is None:
            repeat = self.setting["repeat"]

        self._toggle_repeat(repeat)

    # ==================================================
    def _set_clip(self, clip=None):
        if clip is None:
            clip = self.setting["clip"]

        self._toggle_clip(clip)

    # ==================================================
    def _close_dialog(self):
        if self._dialog_dataset:
            self.tab.close()
            self._dialog_dataset = False
            self.set_status("")
            self._garbage_collection()

    # ==================================================
    def _view_dataset(self):
        def spot(n, r):
            if n != "caption":
                self._plot_spotlight(n, r)

        if not self._dialog_dataset:
            self._dialog_dataset = True
            self.tab = GroupTab(ds=self.dataset, title="DataSet - " + self.setting["model"], parent=self)
            self.tab.rawDataChanged.connect(self._plot_object)
            self.tab.textMessage.connect(self.set_status)
            self.tab.selectedData.connect(spot)
            self.tab.deselectedData.connect(self._remove_spotlight)
            self.tab.finished["int"].connect(self._close_dialog)
            self.tab.show()

    # ==================================================
    def _load(self):
        default_ext = rcParams["plotter.ext"]
        loadable_ext = rcParams["plotter.cif"] + " " + rcParams["plotter.vesta"]
        default_file = os.getcwd()
        ext_str = default_ext + " " + loadable_ext
        filename, _ = QFileDialog.getOpenFileName(
            self, "Load QtDraw", default_file, "QtDraw file (" + ext_str.replace(".", "*.") + ")"
        )
        if filename:
            _, ext = os.path.splitext(filename)
            if ext == "":
                ext = default_ext
                filename += ext
            if ext in ext_str.split(" "):
                self._close_dialog()
                self._load_file(filename, ext)
                self.set_status(f"loaded {filename}.")
            else:
                self.set_status(f"{ext} is unsupported, use {default_ext}.")

    # ==================================================
    def _load_file(self, filename, ext):
        self._homedir = os.path.dirname(filename)
        os.chdir(self._homedir)

        if ext == rcParams["plotter.cif"]:
            plot_cif(self, filename)
            return
        elif ext == rcParams["plotter.vesta"]:
            plot_vesta(self, filename)
            return

        load_dict = read_dict(filename)

        self.preference = load_dict["preference"]
        self.setting = load_dict["setting"]
        self._remove_all_actor()
        self._init_all()
        self.set_view()

        if self.setting["cluster"]:
            self.button_repeat.hide()

        for key in self.dataset.keys():
            if key in load_dict.keys():
                for item in load_dict[key]:
                    self.dataset.append(key, item)
        self._plot_all_object()
        self._layer.set_viewup(self.setting["camera.up"])
        self._layer.set_focus(self.setting["camera.focus"])
        self._layer.set_position(self.setting["camera.position"], reset=True)

        if "multipie" in load_dict.keys():
            if not hasattr(self, "group_panel"):
                self._multipie()
            self.group_panel.load_dict(load_dict["multipie"])

    # ==================================================
    def _save(self):
        default_ext = rcParams["plotter.ext"]
        default_file = os.getcwd() + "/" + self.setting["model"] + default_ext
        default_file = default_file.replace(os.sep, "/")
        filename, _ = QFileDialog.getSaveFileName(self, "Save QtDraw", default_file, "QtDraw file (*" + default_ext + ")")
        if filename:
            _, ext = os.path.splitext(filename)
            if ext == "":
                ext = default_ext
                filename += ext
            if ext == default_ext:
                self._close_dialog()
                self._save_file(filename)
                self._homedir = os.path.dirname(filename)
                os.chdir(self._homedir)
                self.set_status(f"saved: {filename}.")
            else:
                self.set_status(f"{ext} is unsupported, use {default_ext}.")

    # ==================================================
    def create_dict(self, name):
        """
        create model dict for save.

        Args:
            name (str): model name.

        Returns:
            dict: model dict for save.
        """
        self._set_model(name)
        self.setting["axis_mode"] = self.combo_axis.currentIndex()
        self.setting["cell_mode"] = self.combo_cell.currentIndex()
        self.setting["camera.position"] = self._layer.camera.position
        self.setting["camera.up"] = self._layer.camera.up
        self.setting["camera.focus"] = self._layer.camera.focal_point

        save_dict = {"version": __version__, "setting": self.setting, "preference": self.preference}

        self._garbage_collection()

        for key, df in self.dataset.items():
            df0 = df.copy()
            if key in ["caption", "text"]:
                df0.loc[:, "AD"] = ""
            else:
                df0.loc[:, "AD"] = ""
                df0.loc[:, "AL"] = ""
            save_dict[key] = df0.to_dict(orient="split")["data"]

        if hasattr(self, "group_panel"):
            save_dict["multipie"] = self.group_panel.save_dict()

        return save_dict

    # ==================================================
    def _save_file(self, filename):
        base = os.path.splitext(os.path.basename(filename))[0]
        save_dict = self.create_dict(base)
        write_dict(filename, save_dict, "\nQtDraw data file in Python dict format.\n")

    # ==================================================
    def _clear(self):
        ret = QMessageBox.question(self, "", "Are you sure ?", QMessageBox.Ok, QMessageBox.Cancel)
        if ret == QMessageBox.Ok:
            self._close_dialog()
            self._remove_all_actor()
            self._init_dataset()
            self._init_counter
            self._init_actor()
            self._init_setting(
                title=self.setting["title"],
                size=self.setting["size"],
                axis_type=self.preference["axis.type"],
                clip=self.setting["clip"],
                cluster=self.setting["cluster"],
            )
            self._init_panel()
            self._set_light()
            self.set_view()
            self.set_status("removed all objects.")

    # ==================================================
    def _garbage_collection(self):
        for df in self.dataset.values():
            idx = df.index[df.iloc[:, 0] == ""]
            df.drop(idx, inplace=True)
            df.reset_index(drop=True, inplace=True)

    # ==================================================
    def _actor_id(self):
        c = self._counter["actor"]
        s = f"L{c}"
        self._counter["actor"] = c + 1

        return s

    # ==================================================
    def _show_actor(self, name, show):
        self._actorset[name].SetVisibility(show)

    # ==================================================
    def _remove_actor(self, name):
        self._screen_off()
        if name in self._actorset.keys():
            obj_actor = self._actorset[name]
            self._layer.remove_actor(obj_actor)
            del self._actorset[name]
        self._screen_on()

    # ==================================================
    def _remove_all_actor(self):
        names = list(self._actorset.keys())
        for name in names:
            self._remove_actor(name)

    # ==================================================
    def _get_name(self, name):
        c = self._counter[name]
        s = f"{self.def_value[name][0]}{c}"
        self._counter[name] = c + 1
        return s

    # ==================================================
    def _hide_all_object(self):
        """
        hide all object.
        """
        for i in self._actorset.values():
            i.SetVisibility(False)

    # ==================================================
    def _show_all_object(self):
        """
        show all object.
        """
        for group in self.dataset.keys():
            for row_idx in range(len(self.dataset[group])):
                self._show_object(group, row_idx)

    # ==================================================
    def _show_object(self, group, row_idx):
        """
        show object.

        Args:
            group (str): type of object.
            row_idx (int): row index of dataframe to show.
        """
        row = self.dataset[group].iloc[row_idx, :]

        if row[0] == "":
            return

        if group in ["caption", "text"]:
            obj_show, obj_id = row[-2:]
        else:
            obj_show, obj_id, lbl_id = row[-3:]
            lbl_show = row[1]
            self._show_actor(lbl_id, lbl_show)
        self._show_actor(obj_id, obj_show)

    # ==================================================
    def _plot_spotlight(self, group, row_idx):
        size = self.preference["spotlight.size"]
        size *= self._volume ** (1 / 3)
        opt = {
            "color": all_colors[self.preference["spotlight.color"]][0],  # hex
            "opacity": self.preference["spotlight.opacity"],
            "smooth_shading": self.shading,
            "name": f"spotlight_{row_idx}",
        }
        if group == "text":
            return

        th_phi = (rcParams["plot.orbital.theta"], rcParams["plot.orbital.phi"])

        cell = self.dataset[group].loc[row_idx, "cell"]
        position = self.dataset[group].loc[row_idx, "position"]
        cell = NSArray(cell, "vector", "value").astype(int)
        position = NSArray(position, "vector", "value")
        pos = (cell + position).transform(self._A)
        obj = create_sphere(size, th_phi, (0, 180), (0, 360)).translate(pos, inplace=True)
        actor = self._layer.add_mesh(obj, **opt)
        self._spotlight_actor.append((actor, group, row_idx))

    # ==================================================
    def _replot_spotlight(self):
        sp = self._spotlight_actor.copy()
        self._spotlight_actor = []
        for i in sp:
            self._plot_spotlight(i[1], i[2])

    # ==================================================
    def _remove_spotlight(self):
        for i in self._spotlight_actor:
            self._layer.remove_actor(i[0])
        self._spotlight_actor = []

    # ==================================================
    def _plot_to_data(self, group, name, position, info, label="", show_lbl=False, cell=[0, 0, 0]):
        """
        store plot data.

        Args:
            group (str): type of object.
            name (str): name of object.
            position (list): position of object (reduced), [x, y, z].
            info (list): other options such as size, opacity etc.
            label (str, optional): label of object.
            show_lbl (bool, optional): show label ?
            cell (list, optional): cell position.
        """
        if name is None:  # default name.
            name = self._get_name(group)

        if type(cell) == str:
            cell = NSArray(cell).tolist()

        if group == "caption":
            info1 = [name, cell, position] + info + [True, ""]
            self.dataset.append(group, info1)
        elif group == "text":
            info1 = [name, position] + info + [True, ""]
            self.dataset.append(group, info1)
        else:
            if label == "":
                label = name
            for i in position:
                info1 = [name, show_lbl, label, cell, i] + info + [True, "", ""]
                self.dataset.append(group, info1)

    # ==================================================
    def _plot_all_object(self):
        """
        plot all object.
        """
        for group in self.dataset.keys():
            for row_idx in range(len(self.dataset[group])):
                self._plot_object(group, row_idx)

    # ==================================================
    def _plot_object(self, group, row_idx):
        """
        plot object.

        Args:
            group (str): type of object.
            row_idx (int): row index of dataframe to plot.
        """
        row = self.dataset[group].iloc[row_idx, :]

        if row[0] == "":
            if group in ["caption", "text"]:
                obj_id = row[-1]
            else:
                obj_id, lbl_id = row[-2:]
            if obj_id != "":
                self._remove_actor(obj_id)
            if group not in ["caption", "text"] and lbl_id != "":
                self._remove_actor(lbl_id)
            return

        self._screen_off()
        if group == "caption":
            opt, obj_id, obj_show = self._data_to_object(group, row)
            if obj_id == "":
                obj_id = self._actor_id()
                self.dataset[group].iat[row_idx, -1] = obj_id
            obj_actor = self._layer.add_point_labels(name=obj_id, **opt)
            self._actorset[obj_id] = obj_actor
            self._show_actor(obj_id, obj_show)
        elif group == "text":
            opt, obj_id, obj_show = self._data_to_object(group, row)
            if obj_id == "":
                obj_id = self._actor_id()
                self.dataset[group].iat[row_idx, -1] = obj_id
            obj_actor = self._layer.add_text(name=obj_id, **opt)
            self._actorset[obj_id] = obj_actor
            self._show_actor(obj_id, obj_show)
        else:
            opt, obj_id, obj_show, lbl_opt, lbl_id, lbl_show = self._data_to_object(group, row)
            if obj_id == "":
                obj_id = self._actor_id()
                self.dataset[group].iat[row_idx, -2] = obj_id
            obj_actor = self._layer.add_mesh(name=obj_id, **opt)
            self._actorset[obj_id] = obj_actor
            self._show_actor(obj_id, obj_show)

            if lbl_id == "":
                lbl_id = self._actor_id()
                self.dataset[group].iat[row_idx, -1] = lbl_id
            lbl_actor = self._layer.add_point_labels(name=lbl_id, **lbl_opt)
            self._actorset[lbl_id] = lbl_actor
            self._show_actor(lbl_id, lbl_show)
        self._screen_on()

    # ==================================================
    def _check_in_view_range(self, obj):
        """
        check if object is in view range.

        Args:
            obj (NSArray): object.

        Returns:
            bool: True if in range.
        """
        if not self.setting["clip"]:
            return True

        sft = NSArray("[0.001,0.001,0.001]", "vector", "value")
        lower, upper = self.setting["view_range"]
        lower = (NSArray(lower, "vector", "value") - sft).tolist()
        upper = (NSArray(upper, "vector", "value") + sft).tolist()
        idx = obj.clip(lower, upper)

        return len(idx) > 0

    # ==================================================
    def _data_to_object(self, group, row):
        """
        convert string data to object.

        Args:
            group (str): type of object.
            row (list): row of data frame.

        Returns: tuple:
            - (option, object_ID, object_show) for caption and text.
            - (option, object_ID, object_show, label_option, label_ID, label_show) for others.
        """
        if group == "caption":
            igrid, position = row[1:3]
            info = row[3:-2]
            obj_show, obj_id = row[-2:]
        elif group == "text":
            position = eval(str(row[1]))
            info = row[2:-2]
            obj_show, obj_id = row[-2:]
        else:
            label, igrid, position = row[2:5]
            info = row[5:-3]
            obj_show, obj_id, lbl_id = row[-3:]
            lbl_show = row[1]

        if group == "text":
            position = NSArray(position, "scalar", "value")
        else:
            igrid = NSArray(igrid, "vector", "value").astype(int)
            t = f"[[1,0,0,{igrid[0]}],[0,1,0,{igrid[1]}],[0,0,1,{igrid[2]}],[0,0,0,1]]"
            t = NSArray(t, "matrix", "value")
            position = NSArray(position, "vector", "value")
            pos = position.transform(t)
            pos = pos.transform(self._A)

        if group == "caption":
            caption, space, size, bold, color = info
            space = int(space)
            size = int(size)
            bold = bool(bold)
            color = all_colors[color][0]  # hex

            caption = [" " * space + i for i in caption]

            opt = {
                "points": pos.numpy(),
                "labels": caption,
                "font_size": size,
                "bold": bold,
                "text_color": color,
                "shape": None,
                "show_points": False,
                "always_visible": True,
            }

            return opt, obj_id, obj_show

        if group == "text":
            relative, caption, size, color, font = info
            size = int(size)
            color = all_colors[color][0]  # hex
            font = rcParams["detail.dataset.property"]["text"][1][6][1][int(font)]
            relative = bool(relative)

            opt = {
                "position": position.numpy(),
                "text": caption,
                "font_size": size,
                "color": color,
                "font": font,
                "viewport": relative,
            }

            return opt, obj_id, obj_show

        if not self._check_in_view_range(igrid + position):
            obj_show = False
            lbl_show = False

        if group == "site":
            size, color, opacity, space = info
            size = float(size)
            opacity = float(opacity)
            space = int(space)
            color = all_colors[color][0]  # hex

            th_phi = (rcParams["plot.orbital.theta"], rcParams["plot.orbital.phi"])

            obj = create_sphere(0.07 * size, th_phi, (0, 180), (0, 360))
            opt = {"color": color, "opacity": opacity}

        elif group == "bond":
            v, width, color, color2, opacity, space = info
            v = NSArray(v, "vector", "value")
            vec = v.transform(self._A)
            width = float(width)
            opacity = float(opacity)
            space = int(space)

            if color2 != color:
                obj = create_bond(vec, width, True)
                cmap = custom_colormap([color, color2])
                opt = {"show_scalar_bar": False, "cmap": cmap, "opacity": opacity}
            else:
                obj = create_bond(vec, width, False)
                opt = {"color": all_colors[color][0], "opacity": opacity}

            if not (self._check_in_view_range(igrid + position - v / 2) and self._check_in_view_range(igrid + position + v / 2)):
                obj_show = False
                lbl_show = False

        elif group == "vector":
            v, length, width, offset, color, opacity, space = info
            v = NSArray(v, "vector", "value")
            length = float(length)
            width = float(width)
            offset = float(offset)
            color = all_colors[color][0]  # hex
            opacity = float(opacity)
            space = int(space)

            tip_l = self.preference["vector.tip.length"]
            shaft_r = self.preference["vector.shaft.radius"]
            tip_r = self.preference["vector.tip.radius"]

            obj = create_vector(v, length, width, offset, tip_l, shaft_r, tip_r)
            opt = {"color": color, "opacity": opacity}

        elif group == "orbital":
            shape, surface, size, scale, th0, th1, phi0, phi1, color, opacity, space = info
            size = float(size)
            scale = bool(scale)
            opacity = float(opacity)
            space = int(space)
            th_range = [int(th0), int(th1)]
            phi_range = [int(phi0), int(phi1)]

            th_phi = (rcParams["plot.orbital.theta"], rcParams["plot.orbital.phi"])

            obj = create_orbital(shape, surface, size, th_phi, scale, th_range, phi_range)
            if check_color(color):
                opt = {"color": all_colors[color][0], "opacity": opacity}
            else:
                opt = {
                    "cmap": color.strip("*"),
                    "scalars": "surface",
                    "clim": [-1, 1],
                    "opacity": opacity,
                    "show_scalar_bar": False,
                }

        elif group == "stream":
            shape, vector, size, v_size, width, scale, theta, phi, th0, th1, phi0, phi1, color, component, opacity, space = info
            size = float(size)
            v_size = float(v_size)
            width = float(width)
            scale = bool(scale)
            theta = int(theta)
            phi = int(phi)
            th_range = [int(th0), int(th1)]
            phi_range = [int(phi0), int(phi1)]
            if component == 3:
                component = None
            opacity = float(opacity)
            space = int(space)

            tip_l = self.preference["vector.tip.length"]
            shaft_r = self.preference["vector.shaft.radius"]
            tip_r = self.preference["vector.tip.radius"]

            v = create_stream_vector(shape, vector, size, (theta, phi), th_range, phi_range)
            g = create_vector("[1, 0, 0]", 1.0, width, -0.43, tip_l, shaft_r, tip_r)
            if scale:
                obj = v.glyph(orient="vector", scale="vector_abs", factor=v_size, geom=g)
            else:
                obj = v.glyph(orient="vector", scale=None, factor=v_size, geom=g)

            if check_color(color):
                opt = {"clim": [-1, 1], "color": all_colors[color][0], "opacity": opacity}
            else:
                opt = {
                    "clim": [-1, 1],
                    "scalars": "GlyphVector",
                    "cmap": color.strip("*"),
                    "show_scalar_bar": False,
                    "component": component,
                    "opacity": opacity,
                }

        elif group == "plane":
            v, x, y, color, opacity, space = info
            v = NSArray(v, "vector", "value")
            vec = v.transform(self._A)
            x = float(x)
            y = float(y)
            color = all_colors[color][0]  # hex
            opacity = float(opacity)
            space = int(space)
            obj = create_plane(vec, x, y)
            opt = {"color": color, "opacity": opacity}

        elif group == "box":
            a1, a2, a3, edge, wireframe, width, color, opacity, space = info
            a1 = NSArray(a1, "vector", "value")
            a2 = NSArray(a2, "vector", "value")
            a3 = NSArray(a3, "vector", "value")
            a1A = a1.transform(self._A)
            a2A = a2.transform(self._A)
            a3A = a3.transform(self._A)
            edge = bool(edge)
            wireframe = bool(wireframe)
            width = float(width)
            color = all_colors[color][0]  # hex
            opacity = float(opacity)
            space = int(space)
            obj = create_box(a1A, a2A, a3A)
            if wireframe:
                opt = {"color": color, "opacity": opacity, "style": "wireframe", "line_width": width}
            else:
                opt = {"color": color, "opacity": opacity, "show_edges": edge, "line_width": width}

        elif group == "polygon":
            point, cnt, edge, wireframe, width, color, opacity, space = info
            point = NSArray(point, "vector", "value")
            pointA = point.transform(self._A)
            edge = bool(edge)
            wireframe = bool(wireframe)
            width = float(width)
            color = all_colors[color][0]  # hex
            opacity = float(opacity)
            space = int(space)
            obj = create_polygon(pointA, cnt)
            if wireframe:
                opt = {"color": color, "opacity": opacity, "style": "wireframe", "line_width": width}
            else:
                opt = {"color": color, "opacity": opacity, "show_edges": edge, "line_width": width}

        elif group == "text3d":
            text, size, depth, n, offset, color, opacity, space = info
            size = float(size)
            depth = float(depth)
            n = NSArray(n, "vector", "value")
            offset = NSArray(offset, "vector", "value")
            color = all_colors[color][0]  # hex
            opacity = float(opacity)
            space = int(space)
            obj = create_text(text, 0.04 * size, n, self._A_norm, offset, 0.3 * depth)
            opt = {"color": color, "opacity": opacity}

        elif group == "spline":
            point, width, n_interp, closed, natural, color, opacity, space = info
            point = NSArray(point, "vector", "value")
            pointA = point.transform(self._A)
            width = float(width)
            n_interp = int(n_interp)
            closed = bool(closed)
            natural = bool(natural)
            color = all_colors[color][0]
            opacity = float(opacity)
            space = int(space)
            obj = create_spline(pointA, n_interp, width, closed, natural)
            opt = {"color": color, "opacity": opacity}

        elif group == "spline_t":
            expression, t_range, width, n_interp, closed, natural, color, opacity, space = info
            expression = NSArray(expression, "vector")
            t_range = NSArray(t_range, "vector", "value")
            f = expression.lambdify()
            tp = np.arange(t_range[0], t_range[1], t_range[2])
            point = f(tp)
            pointA = point.transform(self._A)
            width = float(width)
            n_interp = int(n_interp)
            closed = bool(closed)
            natural = bool(natural)
            color = all_colors[color][0]
            opacity = float(opacity)
            space = int(space)
            obj = create_spline(pointA, n_interp, width, closed, natural)
            opt = {"color": color, "opacity": opacity}

        else:
            raise ValueError(f"unknown group {group} is given.")

        pos = (igrid + position).transform(self._A)
        g = obj.copy()
        g = g.translate(pos, inplace=True)
        opt["mesh"] = g
        opt["smooth_shading"] = self.shading

        opt["pbr"] = self.preference["light.pbr"]
        opt["metallic"] = self.preference["light.metallic"]
        opt["roughness"] = self.preference["light.roughness"]

        lbl_opt = {
            "points": pos.numpy(),
            "labels": [" " * space + label],
            "font_family": self.preference["label.font"],
            "font_size": self.preference["label.size"],
            "bold": self.preference["label.bold"],
            "italic": self.preference["label.italic"],
            "text_color": all_colors[self.preference["label.color"]][0],
            "shape": None,
            "show_points": False,
            "always_visible": True,
        }
        return opt, obj_id, obj_show, lbl_opt, lbl_id, lbl_show

    # ==================================================
    def _repeat_data(self):
        # create single cell.
        dataset0 = {}
        for group, df in self.dataset.items():
            if len(df) > 0:
                if group != "text":
                    df0 = df[df["cell"].apply(lambda x: str(x)) == "[0, 0, 0]"].copy().reset_index(drop=True)
                    for i in range(len(df0)):
                        df0.at[i, "cell"] = [0, 0, 0]
                else:
                    df0 = df.copy()
                if group in ["caption", "text"]:
                    df0.loc[:, "AD"] = ""
                else:
                    df0.loc[:, "AD"] = ""
                    df0.loc[:, "AL"] = ""
                dataset0[group] = df0
        self._remove_all_actor()
        self._init_dataset()
        self._init_actor()

        # repeat cells.
        if self.setting["repeat"]:
            for group, df in dataset0.items():
                if group == "text":
                    self.dataset[group] = df.reset_index(drop=True)
                else:
                    for grid in self._igrid:
                        df0 = df.reset_index(drop=True)
                        for i in range(len(df0)):
                            df0.at[i, "cell"] = grid
                        self.dataset[group] = pd.concat([self.dataset[group], df0], ignore_index=True, axis=0)
        else:
            for group, df in dataset0.items():
                self.dataset[group] = df.reset_index(drop=True)

        self._plot_all_object()

        if self.setting["repeat"]:
            self.set_status("repeated all objects.")
        else:
            self.set_status("unrepeated all objects.")

    # ==================================================
    def _set_preference(self):
        def apply():
            self._set_light()
            self.set_crystal(self.setting["crystal"])
            self._draw_axis()
            self._draw_cell()
            self._plot_all_object()
            self._replot_spotlight()

        preference = DialogPreference(self.preference, self.setting, apply, self)
        preference.show()

    # ==================================================
    def _about(self):
        about = DialogAbout(self)
        about.exec()

    # ==================================================
    def _multipie(self):
        """
        MultiPie extension.
        """
        from multipie import get_binary

        self.button_clear.hide()
        self.button_multipie.hide()
        self.layout_dataset.removeWidget(self.button_clear)
        self.preference["axis.type"] = "abc"
        self._draw_axis()
        self._core = get_binary(load=True)
        self._create_group_panel()

    # ==================================================
    def _create_group_panel(self):
        from qtdraw.multipie.dialog_group import DialogGroup

        self.group_panel = DialogGroup(core=self._core, qtdraw=self, parent=self)
        self.group_panel.force_close = False
        self.group_panel.show()

    # ==================================================
    def closeEvent(self, event):
        if self.background:
            self.close()
            return
        ret = QMessageBox.question(self, "", "Quit QtDraw ?", QMessageBox.Ok, QMessageBox.Cancel)
        if ret != QMessageBox.Ok:
            event.ignore()
        else:
            if hasattr(self, "group_panel"):
                self.group_panel.force_close = True
                self.group_panel.close()
            self.close()

    # ==================================================
    def close(self):
        super().close()


# ==================================================
class QtDraw(QtDrawWidget):
    """
    Draw application.
    """

    # ==================================================
    def __init__(
        self,
        title=None,
        model=None,
        cell=None,
        origin=None,
        view=None,
        size=None,
        axis_type=None,
        view_range=None,
        repeat=False,
        clip=True,
        cluster=False,
        background=False,
        filename=None,
    ):
        """
        initialize the class.

        Args:
            title (str, optional): window title of plotter.
            model (str, optioanl): name of model.
            cell (dict or str, optional): cell information or "hexagonal/trigonal".
            origin (list, optional): cell origin.
            view (list, optional): view with respect to the axes (a,b,c).
            size (tuple, optional): window size, (width, height).
            axis_type (str, optional): axis type, "xyz/abc/abc*".
            view_range (list, optional): display range ([x0,y0,z0],[x1,y1,z1]) (reduced).
            repeat (bool, optional): repeat plot ?
            clip (bool, optional): clip objects out of view_range ?
            cluster (bool, optional): cluster without repeat ?
            background (bool, optional): background run ?
            filename (str, optional): if not None, load filename.
        """
        style = rcParams["plotter.style"]
        font = rcParams["plotter.font.family"]
        font_size = rcParams["plotter.font.size"]
        self.app = create_application(style=style, font=font, font_size=font_size)
        super().__init__(title, model, cell, origin, view, size, axis_type, view_range, repeat, clip, cluster, background)
        if not check_latex_installed():
            QMessageBox.critical(None, "Error", f"LaTeX command, '{latex_cmd}', cannot be found.", QMessageBox.Yes)
            exit()
        if filename is not None and os.path.isfile(filename):
            _, ext = os.path.splitext(filename)
            self._load_file(filename, ext)
            self.set_status(f"loaded {filename}.")

    # ==================================================
    def show(self):
        """
        show the plot.
        """
        super().show()
        self.app.exec()
        del self.app


# ==================================================
class QtMultiDraw(dict):
    """
    MultiPlotter application.
    """

    # ==================================================
    def __init__(
        self,
        n=1,
        title=None,
        model=None,
        cell=None,
        origin=None,
        view=None,
        size=None,
        axis_type=None,
        view_range=None,
        repeat=None,
        clip=None,
        cluster=None,
        background=None,
    ):
        """
        initialize the class.

        Args:
            n (int, optional): number of plotters.
            title ([str], optional): window title of each plotter.
            model ([str], optioanl): name of model of each plotter.
            cell ([dict or str], optional): cell information or "hexagonal/trigonal" of each plotter.
            origin ([list], optional): cell origin of each plotter.
            view ([list], optional): view with respect to the axes (a,b,c) of each plotter.
            size ([tuple], optional): window size, (width, height) of each plotter.
            axis_type ([str], optional): axis type, "xyz/abc/abc*" of each plotter.
            view_range ([list], optional): display range ([x0,y0,z0],[x1,y1,z1]) (reduced) of each plotter.
            repeat ([bool], optional): repeat plot of each plotter ?
            clip (bool, optional): clip objects out of view_range ?
            cluster (bool, optional): cluster without repeat ?
            background (bool, optional): background run ?

        Notes:
            - number of list content must be equal to n.
        """
        style = rcParams["plotter.style"]
        font = rcParams["plotter.font.family"]
        font_size = rcParams["plotter.font.size"]
        self.app = create_application(style=style, font=font, font_size=font_size)
        if not check_latex_installed():
            QMessageBox.critical(None, "Error", f"LaTeX command, '{latex_cmd}', cannot be found.", QMessageBox.Yes)
            exit()

        if title is None:
            t = rcParams["init.title"]
            title = [f"{t} {i}" for i in range(n)]
        if model is None:
            model = [f"untitled_{i}" for i in range(n)]
        if cell is None:
            cell = [None] * n
        if origin is None:
            origin = [None] * n
        if view is None:
            view = [None] * n
        if size is None:
            size = [None] * n
        if axis_type is None:
            axis_type = [None] * n
        if view_range is None:
            view_range = [None] * n
        if repeat is None:
            repeat = [False] * n
        if clip is None:
            clip = [True] * n
        if cluster is None:
            cluster = [False] * n
        if background is None:
            background = [False] * n

        for i in range(n):
            self[i] = QtDrawWidget(
                title=title[i],
                model=model[i],
                cell=cell[i],
                origin=origin[i],
                view=view[i],
                size=size[i],
                axis_type=axis_type[i],
                view_range=view_range[i],
                repeat=repeat[i],
                clip=clip[i],
                cluster=cluster[i],
                background=background[i],
            )

    # ==================================================
    def show(self):
        """
        show all plotter.
        """
        for p in self.values():
            p.show()
        self.app.exec()
        del self.app



from qtpy.QtWidgets import QLineEdit, QWidget, QLabel, QDialog, QGridLayout
from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QValidator
from gcoreutils.convert_util import is_valid_sympy, text_to_list
from gcoreutils.basic_util import apply
from gcoreutils.string_util import remove_space
from qtdraw.core.validator import validator_map, create_validator


# ==================================================
class LineEdit(QLineEdit):
    """
    customized line editor.
    """

    closeEdit = Signal()
    _valid_style = "padding-left: 3px; background: none;"
    _invalid_style = "padding-left: 3px; background: pink;"
    _read_only_style = "padding-left: 3px; background: lightgray;"
    _validator_list = list(validator_map.keys())

    # ==================================================
    def __init__(self, init="", validator=None, decimal=4, check_var=None, parent=None):
        """
        initialize the class.

        Args:
            init (str, optional): initial text.
            validator (str, optional): validator. see, validator_list.
            decimal (int, optional): precision for real number.
            check_var (list, optional): acceptable variable strings. if None, do not check.
            parent (QWidget, optional): parent object.
        """
        super().__init__(parent)

        self.set_validator(validator, decimal, check_var)
        self.setText(init)
        self._backtup = self.text()

    # ==================================================
    def set_validator(self, validator=None, decimal=4, check_var=None):
        """
        set validator.

        Args:
            validator (str, optional): validator. see, validator_list.
            decimal (int, optional): precision for real number.
            check_var (list, optional): acceptable variable strings. if None, do not check.
        """
        if validator not in self.validator_list:
            validator = None

        self._formatter = lambda text: text

        self._str_validator = False
        if validator is None:
            self._validator = None
        else:
            self._validator = create_validator(validator)
            if validator == "real_unit":
                self._validator.setDecimals(decimal)
            if validator[0] == "i":
                self._formatter = lambda text: str(apply(lambda t: "{ts: d}".format(ts=int(t)), text_to_list(text))).replace(
                    "'", ""
                )
            elif validator[0] == "r":
                self._formatter = lambda text: str(
                    apply(lambda t: "{ts: .{decimal}f}".format(ts=float(t), decimal=decimal), text_to_list(text))
                ).replace("'", "")
            elif validator[0] in ["s", "l"]:
                self._str_validator = True

        self.check_var = check_var

    # ==================================================
    def focusOutEvent(self, event):
        """
        :meta private:
        """
        self.setStyleSheet(self._valid_style)
        self.closeEdit.emit()
        super().focusOutEvent(event)

    # ==================================================
    def keyPressEvent(self, event):
        """
        :meta private:
        """
        k = event.key()
        if k == Qt.Key_Escape:
            self.setStyleSheet(self._valid_style)
            self.closeEdit.emit()
        elif k == Qt.Key_Return:
            status = self.setText(self.text())
            if status:
                self.returnPressed.emit()
        else:
            super().keyPressEvent(event)

    # ==================================================
    def setText(self, text):
        """
        set text.

        Args:
            text (str): text.

        Returns:
            bool: return True if text is valid.
        """
        if self._validator is not None and not self._str_validator:
            text = remove_space(text)
        status = True
        if self._validator is not None:
            status, text, pos = self._validator.validate(text, 0)
            status = status == QValidator.Acceptable
        if self.check_var is not None:
            status = status and is_valid_sympy(text, self.check_var)

        if status:
            text = self._formatter(text)
            self.setStyleSheet(self._valid_style)
            super().setText(text)
            self._backtup = text
        else:
            self.setStyleSheet(self._invalid_style)

        if self.isReadOnly():
            self.setStyleSheet(self._read_only_style)

        return status

    # ==================================================
    def clear(self):
        """
        clear editor.
        """
        self.setText("")
        self.setStyleSheet(self._valid_style)
        super().clear()

    # ==================================================
    @property
    def validator_list(self):
        """
        available validator list.

        Returns:
            list: validator list.
        """
        return self._validator_list


# ==================================================
class LineEditor(QDialog):
    """
    customized line editor widget.
    """

    closeEdit = Signal()
    acceptEdit = Signal()

    # ==================================================
    def __init__(self, label="", init="", validator=None, decimal=4, check_var=None, callback=None, parent=None):
        """
        initialize the class.

        Args:
            label (str, optional): label of the editor.
            init (str, optional): initial text.
            validator (str, optional): validator. see, validator_list.
            decimal (int, optional): precision for real number.
            check_var (list, optional): acceptable variable strings. if None, do not check.
            parent (QWidget, optional): parent object.
        """
        super().__init__(parent)

        if callback is None:
            self.callback = self.default_callback
        else:
            self.callback = callback

        self.label = QLabel(label, self)
        self.label.setIndent(0)

        self.editor = LineEdit(init, validator, decimal, check_var, self)
        self.editor.closeEdit.connect(self.close_edit)
        self.editor.returnPressed.connect(self.accept_edit)

        self.layout = QGridLayout()
        self.layout.setContentsMargins(0, 1, 0, 1)
        self.layout.setSpacing(3)
        if label != "":
            self.layout.addWidget(self.label, 0, 0)
            self.layout.addWidget(self.editor, 0, 1)
        else:
            self.layout.addWidget(self.editor, 0, 0)

        self.setLayout(self.layout)

    # ==================================================
    def close_edit(self):
        """
        :meta private:
        """
        self.editor.setText(self.editor._backtup)
        self.closeEdit.emit()

    # ==================================================
    def accept_edit(self):
        """
        :meta private:
        """
        text = self.editor.text()
        self.callback(text)
        self.acceptEdit.emit()

    # ==================================================
    def setText(self, text):
        """
        set text.

        Args:
            text (str): text.
        """
        self.editor.setText(text)

    # ==================================================
    def text(self):
        """
        current text.

        Returns:
            str: text.
        """
        return self.editor.text()

    # ==================================================
    def default_callback(self, text):
        """
        :meta private:
        """
        print(text)

    # ==================================================
    def set_validator(self, validator=None, decimal=4, check_var=None):
        """
        set validator.

        Args:
            validator (str, optional): validator. see, validator_list.
            decimal (int, optional): precision for real number.
            check_var (list, optional): acceptable variable strings. if None, do not check.
        """
        self.editor.set_validator(validator, decimal, check_var)

    # ==================================================
    def clear(self):
        """
        clear editor.
        """
        self.editor.clear()

    # ==================================================
    @property
    def validator_list(self):
        """
        available validator list.

        Returns:
            list: validator list.
        """
        return self.editor.validator_list


# ==================================================
class Label(QWidget):
    """
    label widget with the same size as editor.
    """

    # ==================================================
    def __init__(self, label="", parent=None):
        """
        initialize the class.

        Args:
            label (str, optional): label.
            parent (QWidget, optional): parent object.
        """
        super().__init__(parent)

        self.label = QLabel(label, self)
        self.label.setIndent(0)

        layout = QGridLayout()
        layout.setContentsMargins(0, 1, 0, 1)
        layout.setSpacing(3)
        layout.addWidget(self.label, 0, 0)

        self.setLayout(layout)

    # ==================================================
    def setText(self, text):
        self.label.setText(text)

    # ==================================================
    def setPixmap(self, pixmap):
        self.label.setPixmap(pixmap)

    # ==================================================
    def setAlignment(self, align):
        if align is None:
            return
        d = {"left": Qt.AlignLeft, "center": Qt.AlignCenter, "right": Qt.AlignRight}
        self.label.setAlignment(d[align] | Qt.AlignVCenter)



import numpy as np
import sympy as sp
import pyvista as pv
from gcoreutils.nsarray import NSArray
from gcoreutils.convert_util import text_to_list, text_to_sympy
from qtdraw.core.util import view_vectors, SplineEx

CHOP = 1e-4


# ==================================================
def str_poly_array(poly, xyz, normalize=True, var=["x", "y", "z"]):
    """
    convert from polynomial string to scalar array.

    Args:
        poly (str): polynomial string of var.
        xyz (ndarray): list of (x,y,z) points (given by create_sphere and .points).
        normalize (bool, optional): normalize ?
        var (list, optional): polynomial variables.

    Returns:
        ndarray: scalar array.
    """
    r = sp.symbols(" ".join(var), real=True)
    v = {var[0]: r[0], var[1]: r[1], var[2]: r[2]}
    poly = poly.replace("sqrt", "SQ")
    poly = poly.replace("r", "(sqrt(x**2+y**2+z**2))")
    poly = poly.replace("SQ", "sqrt")
    ex = text_to_sympy(poly, local=v)
    x, y, z = xyz.T

    f = sp.lambdify(r, ex)
    fv = f(x, y, z)
    if not sorted(map(str, ex.atoms(sp.Symbol))):  # for const.
        fv = np.full(np.size(x), fv)

    if poly != "0" and normalize:
        fv = fv / np.abs(fv).max()

    return fv


# ==================================================
def str_vec_array(vec, xyz, normalize=True, var=["x", "y", "z"]):
    """
    convert from polynomial string (vector) to vector and scalar (abs.) arrays.

    Args:
        vec (NSArray): polynomial component.
        xyz (ndarray): list of (x,y,z) points (given by create_sphere and .points).
        normalize (bool, optional): normalize ?
        var (list, optional): polynomial variables.

    Returns: tuple
        - ndarray: vector array.
        - ndarray: abs. value array.
    """
    f1 = str_poly_array(str(vec[0]), xyz, False, var)
    f2 = str_poly_array(str(vec[1]), xyz, False, var)
    f3 = str_poly_array(str(vec[2]), xyz, False, var)
    f = np.array([f1, f2, f3]).T
    if normalize:
        fa = np.linalg.norm(f, axis=1).max()
        f = f / fa

    fs = np.linalg.norm(f, axis=1)

    return f, fs


# ==================================================
def create_sphere(radius, theta_phi_resolution, theta_range, phi_range):
    """
    create sphere object.

    Args:
        radius (float): radius.
        theta_phi_resolution (list): theta-phi resolution.
        theta_range (list): theta range.
        phi_range (list): phi range.

    Returns:
        PolyData: sphere object.
    """
    th, phi = theta_phi_resolution
    th0, th1 = theta_range
    ph0, ph1 = phi_range
    obj = pv.Sphere(
        radius=radius, theta_resolution=phi, phi_resolution=th, start_phi=th0, end_phi=th1, start_theta=ph0, end_theta=ph1
    )
    return obj


# ==================================================
def create_bond(v, width=1.0, two=True):
    """
    bond object.

    Args:
        v (str or NSArray): bond direction vector (cartesian).
        width (float, optional): width.
        two (bool, optional): twotone color ?

    Returns:
        PolyData: cylinder object
    """
    if not isinstance(v, (str, NSArray)):
        raise KeyError(f"{type(v)} is invalid.")

    if type(v) == str:
        v = NSArray(v, fmt="value")

    length = v.norm()

    if two:
        p = pv.Cylinder(direction=v.numpy(), radius=0.018 * width, height=length / 2)
        m = pv.PolyData([(-v * 0.25).tolist(), (v * 0.25).tolist()])
        m.point_data["scalars"] = [0, 1]
        obj = m.glyph(geom=(p, p), scale=False, rng=(0, 1))
    else:
        p = pv.Cylinder(direction=v.numpy(), radius=0.018 * width, height=length)
        m = pv.PolyData([[0.0, 0.0, 0.0]])
        m.point_data["scalars"] = [0]
        obj = m.glyph(geom=(p), scale=False, rng=(0, 0))

    return obj


# ==================================================
def create_vector(v, length, width, offset, tip_l, shaft_r, tip_r):
    """
    vector object.

    Args:
        v (str or NSArray): vector direction (to be normalized, cartesian).
        length (float): length.
        width (float): width.
        offset (float): offest ratio.

    Returns:
        PolyData: arrow object.
    """
    if not isinstance(v, (str, NSArray)):
        raise KeyError(f"{type(v)} is invalid.")

    if type(v) == str:
        v = NSArray(v, fmt="value")

    v = v.normalize()

    obj = pv.Arrow(
        offset * v.numpy() * length,
        v.numpy(),
        tip_length=tip_l,
        shaft_radius=shaft_r * width / length,
        tip_radius=tip_r * width / length,
        scale=length,
        tip_resolution=100,
        shaft_resolution=100,
    )
    return obj


# ==================================================
def create_orbital(shape, surface, size, theta_phi_res, scale, theta_range, phi_range):
    """
    orbital object.

    Args:
        shape (str): (x,y,z) shape (cartesian).
        surface (str): (x,y,z) surface color (cartesian).
        size (float): size.
        theta_phi_res (tuple): resolution for theta and phi.
        scale (bool): if True, shappe is scaled.
        theta_range (list): theta range.
        phi_range (list): phi range.

    Returns:
        PolyData: orbital object with "surface".
    """
    obj = create_sphere(size, theta_phi_res, theta_range=theta_range, phi_range=phi_range)
    sp = obj.points
    fs = np.abs(str_poly_array(shape, sp, normalize=scale))
    fc = str_poly_array(surface, sp)
    obj.points = np.tile(fs, (3, 1)).T * sp
    obj["surface"] = np.real(fc)

    return obj


# ==================================================
def create_stream_vector(shape, vector, size, theta_phi_res, theta_range, phi_range):
    """
    steam vector object.

    Args:
        shape (str): f(x,y,z) shape (cartesian).
        vector(str or NSArray): "[vx(x,y,z),vy(x,y,z),vz(x,y,z)]" vector string (cartesina).
        size (float): size.
        theta_phi_res (tuple): resolution for theta and phi.
        theta_range (list): theta range.
        phi_range (list): phi range.

    Returns:
        PolyData: orbital object with "vector" and "vector_abs".
    """
    if not isinstance(vector, (str, NSArray)):
        raise KeyError(f"{type(vector)} is invalid.")

    if type(vector) == str:
        vector = NSArray(vector)

    obj = create_orbital(shape, "0", size, theta_phi_res, True, theta_range, phi_range)
    fv, fva = str_vec_array(vector, obj.points)
    obj["vector"] = fv
    obj["vector_abs"] = fva

    return obj


# ==================================================
def create_plane(v, x=1.0, y=1.0):
    """
    plane object.

    Args:
        v (str or NSArray): vector direction (to be normalized, cartesian).
        x (float, optional): x size.
        y (float, optional): y size.

    Returns:
        PolyData: plane object.
    """
    if not isinstance(v, (str, NSArray)):
        raise KeyError(f"{type(v)} is invalid.")

    if type(v) == str:
        v = NSArray(v, fmt="value")

    v = v.normalize()

    obj = pv.Plane(direction=v.numpy(), i_size=x, j_size=y)
    return obj


# ==================================================
def create_box(a1, a2, a3):
    """
    box object.

    Args:
        a1 (str or NSArray): 1st direction (reduced).
        a2 (str or NSArray): 2nd direction (reduced).
        a3 (str or NSArray): 3rd direction (reduced).

    Returns:
        PolyData: cube object.
    """
    if not isinstance(a1, (str, NSArray)):
        raise KeyError(f"{type(a1)} is invalid.")
    if not isinstance(a2, (str, NSArray)):
        raise KeyError(f"{type(a2)} is invalid.")
    if not isinstance(a3, (str, NSArray)):
        raise KeyError(f"{type(a3)} is invalid.")

    if type(a1) == str:
        a1 = NSArray(a1, fmt="value")
    if type(a2) == str:
        a2 = NSArray(a2, fmt="value")
    if type(a3) == str:
        a3 = NSArray(a3, fmt="value")

    A = np.eye(4)
    A[0:3, 0] = a1.numpy()
    A[0:3, 1] = a2.numpy()
    A[0:3, 2] = a3.numpy()

    obj = pv.Cube(bounds=(0.0, 1.0, 0.0, 1.0, 0.0, 1.0))
    obj.transform(A, inplace=True)
    return obj


# ==================================================
def create_linebox(a1, a2, a3):
    """
    linebox object.

    Args:
        a1 (str or NSArray): 1st direction (reduced).
        a2 (str or NSArray): 2nd direction (reduced).
        a3 (str or NSArray): 3rd direction (reduced).

    Returns:
        PolyData: linebox object.
    """
    if not isinstance(a1, (str, NSArray)):
        raise KeyError(f"{type(a1)} is invalid.")
    if not isinstance(a2, (str, NSArray)):
        raise KeyError(f"{type(a2)} is invalid.")
    if not isinstance(a3, (str, NSArray)):
        raise KeyError(f"{type(a3)} is invalid.")

    if type(a1) == str:
        a1 = NSArray(a1, fmt="value")
    if type(a2) == str:
        a2 = NSArray(a2, fmt="value")
    if type(a3) == str:
        a3 = NSArray(a3, fmt="value")

    A = np.eye(4)
    A[0:3, 0] = a1.numpy()
    A[0:3, 1] = a2.numpy()
    A[0:3, 2] = a3.numpy()

    pts = np.array([[0.0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]])
    lines = np.array([5, 0, 1, 2, 3, 0, 5, 4, 5, 6, 7, 4, 2, 0, 4, 2, 1, 5, 2, 2, 6, 2, 3, 7])
    obj = pv.PolyData(pts, lines=lines)
    obj.transform(A, inplace=True)

    return obj


# ==================================================
def create_polygon(pt, cnt):
    """
    polygon object.

    Args:
        pt (str or NSArray): vertices (cartesian).
        cnt (list): list of connected vectices for each plane.

    Returns:
        PolyData: polygon object.
    """
    if not isinstance(pt, (str, NSArray)):
        raise KeyError(f"{type(pt)} is invalid.")

    if type(pt) == str:
        pt = NSArray(pt, fmt="value")

    n = len(pt)
    if type(cnt) is str:
        cnt = text_to_list(cnt)
    cnt = [[len(cell), *[int(i) % n for i in cell]] for cell in cnt]
    cell = np.hstack(cnt)
    obj = pv.PolyData(pt.numpy(), cell)

    return obj


# ==================================================
def create_text(txt, size, normal, A, offset, depth=1.0):
    """
    text3d object.

    Args:
        txt (str): text.
        depth (float, optional): depth.

    Returns:
        PolyData: text3d object.
    """
    v, u = view_vectors(A, normal)
    A = NSArray(
        [
            [u[1] * v[2] - u[2] * v[1], u[0], v[0], offset[0]],
            [u[2] * v[0] - u[0] * v[2], u[1], v[1], offset[1]],
            [u[0] * v[1] - u[1] * v[0], u[2], v[2], offset[2]],
            [0, 0, 0, 1],
        ],
        "matrix",
        "value",
    )

    obj = pv.Text3D(txt, depth)
    obj.scale([size] * 3, inplace=True)
    obj.transform(A, inplace=True)

    return obj


# ==================================================
def create_spline(pts, n_interp, width=1.0, closed=False, natural=True):
    """
    spline object.

    Args:
        pts (str or NSArray): points.
        width (float, optional); width.
        n_interp (int, optional): number of interpolation points.
        closed (bool, optional): closed spline ?
        natural (bool, optional): natural boundary ?

    Returns:
        PolyData: spline object.
    """
    if not isinstance(pts, (str, NSArray)):
        raise KeyError(f"{type(pts)} is invalid.")

    if type(pts) == str:
        pts = NSArray(pts, fmt="value")

    obj = SplineEx(pts.numpy(), n_interp, closed, natural)
    obj.tube(radius=0.01 * width, inplace=True)
    return obj
